#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : ML Studio                                                         #
# Version : 0.1.14                                                            #
# File    : test_activations.py                                               #
# Python  : 3.8.3                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Monday, June 15th 2020, 10:24:08 pm                         #
# Last Modified : Monday, June 15th 2020, 10:24:25 pm                         #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
"""Test Activation Functions."""
import math
import numpy as np
import pytest
from pytest import mark

from scipy.special import expit, softmax
from sklearn.linear_model import LinearRegression as skl_linear_regression
from sklearn.linear_model import LogisticRegression as skl_logistic_regression

from mlstudio.supervised.core.tasks import LinearRegression, LogisticRegression
from mlstudio.supervised.core.tasks import MultinomialLogisticRegression
from mlstudio.utils.data_manager import AddBiasTerm
# --------------------------------------------------------------------------  #
@mark.tasks
@mark.linear_regression
class LinearRegressionTaskTests:

    def test_linear_regression_output(self, make_regression_data):        
        X, y = make_regression_data
        # Obtain sklearn parameters and solution
        skl = skl_linear_regression()
        skl.fit(X,y)
        bias = np.atleast_1d(skl.intercept_)
        coef = skl.coef_    
        theta = np.concatenate((bias, coef))    
        y_pred_skl = skl.predict(X)
        # Add bias term to data
        X = AddBiasTerm().fit_transform(X) 
        # Compute solution
        t = LinearRegression()
        y_pred = t.compute_output(theta, X)        
        assert np.allclose(y_pred_skl, y_pred), "Compute output inaccurate"
        y_pred = t.predict(theta, X)        
        assert np.allclose(y_pred_skl, y_pred), "Predict inaccurate"

# --------------------------------------------------------------------------  #
@mark.tasks
@mark.logistic_regression
class LogisticRegressionTaskTests:

    def test_logistic_regression_output(self, make_classification_data):
        X, y = make_classification_data
        # Obtain sklearn parameters and solution
        skl = skl_logistic_regression(penalty='none')
        skl.fit(X,y)
        bias = np.atleast_1d(skl.intercept_)
        coef = skl.coef_    
        theta = np.concatenate((bias, coef[0]))    
        y_prob_skl = skl.predict_proba(X)[:,1]
        y_pred_skl = skl.predict(X)
        # Add bias term to data
        X = AddBiasTerm().fit_transform(X) 
        # Compute solution
        t = LogisticRegression()
        y_pred = t.compute_output(theta, X)        
        assert np.allclose(y_prob_skl, y_pred), "Compute output inaccurate"
        y_pred = t.predict(theta, X)        
        assert np.allclose(y_pred_skl, y_pred), "Predict inaccurate"

# --------------------------------------------------------------------------  #
@mark.tasks
@mark.softmax_regression
class SoftmaxRegressionTaskTests:

    def test_softmax_regression_output(self, make_multiclass_data):
        X, y = make_multiclass_data
        # Obtain sklearn parameters and solution
        skl = skl_logistic_regression(penalty='none')
        skl.fit(X,y)
        bias = np.atleast_2d(skl.intercept_)
        coef = skl.coef_    
        theta = np.concatenate((bias, coef.T), axis=0)    
        y_prob_skl = skl.predict_proba(X)
        y_pred_skl = skl.predict(X)
        # Add bias term to data
        X = AddBiasTerm().fit_transform(X) 
        # Compute solution
        t = MultinomialLogisticRegression()
        y_pred = t.compute_output(theta, X)        
        assert np.allclose(y_prob_skl, y_pred), "Compute output inaccurate"
        y_pred = t.predict(theta, X)        
        assert np.allclose(y_pred_skl, y_pred), "Predict inaccurate"




#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : MLStudio                                                          #
# Version : 0.1.0                                                             #
# File    : model_validation.py                                               #
# Python  : 3.8.2                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Wednesday, March 18th 2020, 5:21:32 am                      #
# Last Modified : Thursday, March 19th 2020, 7:18:57 pm                       #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
"""Model Validation Plots."""
import math
import numpy as np
import pandas as pd
import plotly
import plotly.express as px
import plotly.graph_objects as go 
import plotly.offline as py
from sklearn.model_selection import ParameterGrid, learning_curve 
from sklearn.model_selection import validation_curve
import statsmodels.api as sm
from statsmodels.nonparametric.smoothers_lowess import lowess

from .base import ModelVisualatrix
from mlstudio.analysis.model_validation import standardized_residuals
from mlstudio.analysis.model_validation import studentized_residuals, quantile
from mlstudio.analysis.model_validation import leverage, cooks_distance
from mlstudio.visual import COLORS
from mlstudio.supervised.machine_learning.linear_regression import LinearRegression
from mlstudio.utils.format import proper        
# --------------------------------------------------------------------------  #
#                              RESIDUALS                                      #
# --------------------------------------------------------------------------  #
class Residuals(ModelVisualatrix):        
    """Plots residuals versus predicted values.

    Parameters
    ----------
    fig : Plotly Figure or FigureWidget
        The plotting object. 

    estimator : MLStudio estimator object.
        The object that implements the 'fit' and 'predict' methods.

    hist : bool, 'density', 'frequency', default: True

        Draw a histogram showing the distribution of the residuals on the 
        right side of the figure. If set to 'density', the probability
        density will be plotted. If set to 'frequency', the frequency will
        be plotted.       
    
    kwargs : dict
        Keyword arguments that are passed to the base class and influence
        the visualization. Optional keyword arguments include:

        =========   ==========================================
        Property    Description
        --------    ------------------------------------------
        height      specify the height of the figure
        width       specify the width of the figure
        title       specify the title of the figure
        template    specify the template for the figure.
        =========   ==========================================    
    
    """

    def __init__(self, estimator, fig=None, hist=True, **kwargs):
        super(Residuals, self).__init__(estimator=estimator,
                                        fig=fig, **kwargs)

        self.hist = hist
        self.title = self.title or str(estimator.description + "<br>Residuals vs. Predicted")

    def fit(self, X, y):
        """Generates the plot.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training vector, where n_samples is the number of samples and
            n_features is the number of features.

        y : array-like, shape (n_samples) or (n_samples, n_features), optional
            Target relative to X for classification or regression;
            None for unsupervised learning.

        """
        # Compute residuals
        self.estimator.fit(X,y)
        y_pred = self.estimator.predict(X)
        residuals = y - y_pred
        # Flatten arrays (just in case)
        y = y.ravel()
        y_pred = y_pred.ravel()
        residuals = residuals.ravel()

        # Create lowess smoothing line
        z1 = lowess(residuals, y_pred, frac=1./3, it=0, is_sorted=False)

        # Create scatterplot traces
        data = [
            go.Scattergl(x=y_pred, y=residuals,
                        mode='markers',
                        marker=dict(color=COLORS['blue'],
                            line=dict(
                                color='white',
                                width=1
                            )),
                        name="Residual Plot",
                        showlegend=False),

            go.Scattergl(x=z1[:,0], y=z1[:,1],
                        mode='lines',
                        marker=dict(color='red'),
                        name="Training Set Lowess",
                        showlegend=False)                                                
        ]

        # Designate Layout
        layout = go.Layout(title=self.title, 
                        title_x=0.5,
                        height=self.height,
                        width=self.width,
                        xaxis_title="Predicted",
                        yaxis_title="Residuals",
                        xaxis=dict(domain=[0,0.85],  zeroline=False),
                        yaxis=dict(domain=[0,0.85],  zeroline=False),
                        xaxis2=dict(domain=[0.85,1], zeroline=False),
                        yaxis2=dict(domain=[0.85,1], zeroline=False),                        
                        showlegend=False,
                        template=self.template)

        # Create figure object
        self.fig = go.Figure(data=data, layout=layout)                        

        # Create and add shapes
        self.fig.add_shape(
            go.layout.Shape(
                type="line",
                x0=np.min(y_pred),
                y0=0,
                x1=np.max(y_pred),
                y1=0,
                line=dict(
                    color=COLORS['green']
                )
            )
        )
        self.fig.update_shapes(dict(xref='x', yref='y'))

        # Specify existence and type of histogram 
        if self.hist is True:
            self.hist = ""
        if self.hist in ["density", ""]:
            self.fig.add_trace(go.Histogram(y=residuals,
                                name="y density train",
                                showlegend=False,
                                xaxis="x2",
                                orientation="h",
                                marker=dict(color=COLORS['green'],
                                    line=dict(
                                        color='white',
                                        width=1
                                    )),
                                histnorm=self.hist))

# --------------------------------------------------------------------------  #
#                          STANDARDIZED RESIDUALS                             #
# --------------------------------------------------------------------------  #
class StandardizedResiduals(ModelVisualatrix):        
    """Standardized Residual plot.

    Renders a standardized residual plot showing the residuals on the vertical 
    axis and the predicted values on the horizontal access.

    Parameters
    ----------
    fig : Plotly Figure or FigureWidget
        The plotting object. 

    estimator : MLStudio estimator object.
        The object that implements the 'fit' and 'predict' methods.

    hist : bool, 'density', 'frequency', default: True

        Draw a histogram showing the distribution of the residuals on the 
        right side of the figure. If set to 'density', the probability
        density will be plotted. If set to 'frequency', the frequency will
        be plotted.     
    
    kwargs : dict
        Keyword arguments that are passed to the base class and influence
        the visualization. Optional keyword arguments include:

        =========   ==========================================
        Property    Description
        --------    ------------------------------------------
        height      specify the height of the figure
        width       specify the width of the figure
        title       specify the title of the figure
        template    specify the template for the figure.
        =========   ==========================================    
    
    """

    def __init__(self, estimator, fig=None, hist=True, **kwargs):
        super(StandardizedResiduals, self).__init__(estimator=estimator,
                                        fig=fig, **kwargs)

        self.hist = hist
        self.title = self.title or str(estimator.description + \
            "<br>Standardized Residuals vs. Predicted")

    def fit(self, X, y):
        """Generates the plot.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training vector, where n_samples is the number of samples and
            n_features is the number of features.

        y : array-like, shape (n_samples) or (n_samples, n_features), optional
            Target relative to X for classification or regression;
            None for unsupervised learning.

        """
        # Compute standardized residuals
        self.estimator.fit(X,y)
        y_pred = self.estimator.predict(X)
        residuals = y - y_pred
        residuals = standardized_residuals(X, y, y_pred)

        # Flatten arrays (just in case)
        y = y.ravel()
        y_pred = y_pred.ravel()
        residuals = residuals.ravel()

        # Create lowess smoothing line
        z1 = lowess(residuals, y_pred, frac=1./3, it=0, is_sorted=False)

        # Create scatterplot traces
        data = [
            go.Scattergl(x=y_pred, y=residuals,
                        mode='markers',
                        marker=dict(color=COLORS['blue'],
                            line=dict(
                                color='white',
                                width=1
                            )),
                        name="Residual Plot",
                        showlegend=False),

            go.Scattergl(x=z1[:,0], y=z1[:,1],
                        mode='lines',
                        marker=dict(color='red'),
                        name="Training Set Lowess",
                        showlegend=False)                                                
        ]

        # Designate Layout
        layout = go.Layout(title=self.title, 
                        title_x=0.5,
                        height=self.height,
                        width=self.width,
                        xaxis_title="Predicted",
                        yaxis_title="Standardized Residuals",
                        xaxis=dict(domain=[0,0.85],  zeroline=False),
                        yaxis=dict(domain=[0,0.85],  zeroline=False),
                        xaxis2=dict(domain=[0.85,1], zeroline=False),
                        yaxis2=dict(domain=[0.85,1], zeroline=False),                        
                        showlegend=False,
                        template=self.template)

        # Create figure object
        self.fig = go.Figure(data=data, layout=layout)                        

        # Create and add shapes
        self.fig.add_shape(
            go.layout.Shape(
                type="line",
                x0=np.min(y_pred),
                y0=0,
                x1=np.max(y_pred),
                y1=0,
                line=dict(
                    color=COLORS['green']
                )
            )
        )
        self.fig.update_shapes(dict(xref='x', yref='y'))

        # Specify existence and type of histogram 
        if self.hist is True:
            self.hist = ""
        if self.hist in ["density", ""]:
            self.fig.add_trace(go.Histogram(y=residuals,
                                name="y density train",
                                showlegend=False,
                                xaxis="x2",
                                orientation="h",
                                marker=dict(color=COLORS['green'],
                                    line=dict(
                                        color='white',
                                        width=1
                                    )),
                                histnorm=self.hist))                                

# --------------------------------------------------------------------------  #
#                          STANDARDIZED RESIDUALS                             #
# --------------------------------------------------------------------------  #
class StudentizedResiduals(ModelVisualatrix):        
    """Studentized Residual plot.

    Renders a studentized residual plot showing the residuals on the vertical 
    axis and the predicted values on the horizontal access.

    Parameters
    ----------
    fig : Plotly Figure or FigureWidget
        The plotting object. 

    estimator : MLStudio estimator object.
        The object that implements the 'fit' and 'predict' methods.

    hist : bool, 'density', 'frequency', default: True

        Draw a histogram showing the distribution of the residuals on the 
        right side of the figure. If set to 'density', the probability
        density will be plotted. If set to 'frequency', the frequency will
        be plotted.     
    
    kwargs : dict
        Keyword arguments that are passed to the base class and influence
        the visualization. Optional keyword arguments include:

        =========   ==========================================
        Property    Description
        --------    ------------------------------------------
        height      specify the height of the figure
        width       specify the width of the figure
        title       specify the title of the figure
        template    specify the template for the figure.
        =========   ==========================================    
    
    """

    def __init__(self, estimator, fig=None, hist=True, **kwargs):
        super(StudentizedResiduals, self).__init__(estimator=estimator,
                                        fig=fig, **kwargs)

        self.hist = hist
        self.title = self.title or str(estimator.description + \
            "<br>Studentized Residuals vs. Predicted")

    def fit(self, X, y):
        """Generates the plot.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training vector, where n_samples is the number of samples and
            n_features is the number of features.

        y : array-like, shape (n_samples) or (n_samples, n_features), optional
            Target relative to X for classification or regression;
            None for unsupervised learning.

        """
        # Compute studentized residuals
        self.estimator.fit(X,y)
        y_pred = self.estimator.predict(X)
        residuals = y - y_pred
        residuals = studentized_residuals(X, y, y_pred)

        # Flatten arrays (just in case)
        y = y.ravel()
        y_pred = y_pred.ravel()
        residuals = residuals.ravel()

        # Create lowess smoothing line
        z1 = lowess(residuals, y_pred, frac=1./3, it=0, is_sorted=False)

        # Create scatterplot traces
        data = [
            go.Scattergl(x=y_pred, y=residuals,
                        mode='markers',
                        marker=dict(color=COLORS['blue'],
                            line=dict(
                                color='white',
                                width=1
                            )),
                        name="Residual Plot",
                        showlegend=False),

            go.Scattergl(x=z1[:,0], y=z1[:,1],
                        mode='lines',
                        marker=dict(color='red'),
                        name="Training Set Lowess",
                        showlegend=False)                                                
        ]

        # Designate Layout
        layout = go.Layout(title=self.title, 
                        title_x=0.5,
                        height=self.height,
                        width=self.width,
                        xaxis_title="Predicted",
                        yaxis_title="Studentized Residuals",
                        xaxis=dict(domain=[0,0.85],  zeroline=False),
                        yaxis=dict(domain=[0,0.85],  zeroline=False),
                        xaxis2=dict(domain=[0.85,1], zeroline=False),
                        yaxis2=dict(domain=[0.85,1], zeroline=False),                        
                        showlegend=False,
                        template=self.template)

        # Create figure object
        self.fig = go.Figure(data=data, layout=layout)                        

        # Create and add shapes
        self.fig.add_shape(
            go.layout.Shape(
                type="line",
                x0=np.min(y_pred),
                y0=0,
                x1=np.max(y_pred),
                y1=0,
                line=dict(
                    color=COLORS['green']
                )
            )
        )
        self.fig.update_shapes(dict(xref='x', yref='y'))

        # Specify existence and type of histogram 
        if self.hist is True:
            self.hist = ""
        if self.hist in ["density", ""]:
            self.fig.add_trace(go.Histogram(y=residuals,
                                name="y density train",
                                showlegend=False,
                                xaxis="x2",
                                orientation="h",
                                marker=dict(color=COLORS['green'],
                                    line=dict(
                                        color='white',
                                        width=1
                                    )),
                                histnorm=self.hist))                                   

# --------------------------------------------------------------------------  #
#                            SCALE LOCATION                                   #
# --------------------------------------------------------------------------  #
class ScaleLocation(ModelVisualatrix):        
    """Scale location plot.

    Renders a scale location plot showing the the square root of the 
    standardized residuals and predicted values. 

    Parameters
    ----------
    fig : Plotly Figure or FigureWidget
        The plotting object. 

    estimator : MLStudio estimator object.
        The object that implements the 'fit' and 'predict' methods.

    hist : bool, 'density', 'frequency', default: True

        Draw a histogram showing the distribution of the residuals on the 
        right side of the figure. If set to 'density', the probability
        density will be plotted. If set to 'frequency', the frequency will
        be plotted.     
    
    kwargs : dict
        Keyword arguments that are passed to the base class and influence
        the visualization. Optional keyword arguments include:

        =========   ==========================================
        Property    Description
        --------    ------------------------------------------
        height      specify the height of the figure
        width       specify the width of the figure
        title       specify the title of the figure
        template    specify the template for the figure.
        =========   ==========================================    
    
    """

    def __init__(self, estimator, fig=None, hist=True, **kwargs):
        super(ScaleLocation, self).__init__(estimator=estimator,
                                        fig=fig, **kwargs)

        self.hist = hist
        self.title = self.title or str(estimator.description + \
            "<br>Scale Location")

    def fit(self, X, y):
        """Generates the plot.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training vector, where n_samples is the number of samples and
            n_features is the number of features.

        y : array-like, shape (n_samples) or (n_samples, n_features), optional
            Target relative to X for classification or regression;
            None for unsupervised learning.

        """
        # Compute standardized residuals
        self.estimator.fit(X,y)
        y_pred = self.estimator.predict(X)
        residuals = y - y_pred
        # Note: residuals parameter is the square root of the standardized residuals.
        residuals = np.sqrt(standardized_residuals(X, y, y_pred))

        # Flatten arrays (just in case)
        y = y.ravel()
        y_pred = y_pred.ravel()
        residuals = residuals.ravel()

        # Create lowess smoothing line
        z1 = lowess(residuals, y_pred, frac=1./3, it=0, is_sorted=False)

        # Create scatterplot traces
        data = [
            go.Scattergl(x=y_pred, y=residuals,
                        mode='markers',
                        marker=dict(color=COLORS['blue'],
                            line=dict(
                                color='white',
                                width=1
                            )),
                        name="Residual Plot",
                        showlegend=False),

            go.Scattergl(x=z1[:,0], y=z1[:,1],
                        mode='lines',
                        marker=dict(color='red'),
                        name="Training Set Lowess",
                        showlegend=False)                                                
        ]

        # Designate Layout
        layout = go.Layout(title=self.title, 
                        title_x=0.5,
                        height=self.height,
                        width=self.width,
                        xaxis_title="Predicted",
                        yaxis_title=r"$\sqrt{\text{Standardized Residuals}}$",
                        xaxis=dict(domain=[0,0.85],  zeroline=False),
                        yaxis=dict(domain=[0,0.85],  zeroline=False),
                        xaxis2=dict(domain=[0.85,1], zeroline=False),
                        yaxis2=dict(domain=[0.85,1], zeroline=False),                        
                        showlegend=False,
                        template=self.template)

        # Create figure object
        self.fig = go.Figure(data=data, layout=layout)                        

        self.fig.update_shapes(dict(xref='x', yref='y'))

        # Specify existence and type of histogram 
        if self.hist is True:
            self.hist = ""
        if self.hist in ["density", ""]:
            self.fig.add_trace(go.Histogram(y=residuals,
                                name="y density train",
                                showlegend=False,
                                xaxis="x2",
                                orientation="h",
                                marker=dict(color=COLORS['green'],
                                    line=dict(
                                        color='white',
                                        width=1
                                    )),
                                histnorm=self.hist))          

# --------------------------------------------------------------------------  #
#                               QQNORM                                        #
# --------------------------------------------------------------------------  #
class QQPlot(ModelVisualatrix):        
    """QQ Plot.

    The quantile-quantile (QQ) plot is used to show if two data sets come from 
    the same distribution. Concretely, one data set's quantiles are plotted 
    along the x-axis and the quantiles for the second distribution are plotted 
    on the y-axis. Typically, data sets are compared to the normal distribution.
    Considered the reference distribution, it is plotted along the X-axis as the
    "Theoretical Quantiles" while the sample is plotted along the Y-axis as
    the "Sample Quantiles".

    Parameters
    ----------
    fig : Plotly Figure or FigureWidget
        The plotting object. 

    estimator : MLStudio estimator object.
        The object that implements the 'fit' and 'predict' methods.

    a : float
        Offset for the plotting position of an expected order
        statistic, for example. The plotting positions are given
        by (i - a)/(nobs - 2*a + 1) for i in range(0,nobs+1)
    
    loc : float
        Location parameter for dist
    
    scale : float
        Scale parameter for dist
    
    kwargs : dict
        Keyword arguments that are passed to the base class and influence
        the visualization. Optional keyword arguments include:

        =========   ==========================================
        Property    Description
        --------    ------------------------------------------
        height      specify the height of the figure
        width       specify the width of the figure
        title       specify the title of the figure
        template    specify the template for the figure.
        =========   ==========================================    
    
    """

    def __init__(self, estimator, a=0, loc=0, scale=1, fig=None, hist=True, 
                 **kwargs):
        super(QQPlot, self).__init__(estimator=estimator,
                                        fig=fig, **kwargs)

        self.hist = hist
        self.a = a
        self.loc = loc
        self.scale = scale
        self.title = self.title or str(estimator.description + \
            "<br>Normal Q-Q")

    def fit(self, X, y):
        """Generates the plot.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training vector, where n_samples is the number of samples and
            n_features is the number of features.

        y : array-like, shape (n_samples) or (n_samples, n_features), optional
            Target relative to X for classification or regression;
            None for unsupervised learning.

        """
        # Compute standardized residuals and sort
        self.estimator.fit(X,y)
        y_pred = self.estimator.predict(X)
        residuals = y - y_pred
        residuals = standardized_residuals(X, y, y_pred)
        residuals = np.sort(residuals)

        # Flatten arrays (just in case)
        y = y.ravel()
        y_pred = y_pred.ravel()        
        residuals = residuals.ravel()

        # Compute theoretical quantiles
        theoretical_quantiles = quantile(residuals)       

        # Create scatterplot traces
        data = [
            go.Scattergl(x=theoretical_quantiles, 
                         y=residuals,
                         mode='markers',
                         marker=dict(color=COLORS['blue']),
                         name="Theoretical Quantiles",
                         showlegend=True),
            go.Scattergl(x=theoretical_quantiles,
                         y=theoretical_quantiles,
                        mode='lines',
                        marker=dict(color='green'),
                        name="Theoretical Quantiles Line",
                        showlegend=False)                            

        ]

        # Designate Layout
        layout = go.Layout(title=self.title, 
                        title_x=0.5,
                        height=self.height,
                        width=self.width,
                        xaxis_title="Theoretical Quantiles",
                        yaxis_title="Standardized Residuals",                       
                        showlegend=False,
                        template=self.template)

        # Create figure object
        self.fig = go.Figure(data=data, layout=layout)        


# --------------------------------------------------------------------------  #
#                         RESIDUALS VS LEVERAGE                               #
# --------------------------------------------------------------------------  #
class ResidualsLeverage(ModelVisualatrix):        
    """Residuals vs Leverage Plot.

    The Residuals vs Leverage plot helps to illuminate data points that may
    be influential. Observations that lie in the upper and lower right 
    corners of the plot may be outside of the Cook's distance line, and 
    would require further investigation.

    Parameters
    ----------
    fig : Plotly Figure or FigureWidget
        The plotting object. 

    estimator : MLStudio estimator object.
        The object that implements the 'fit' and 'predict' methods.

    inner_threshold : float (Default = 0.5)
        The threshold for the cooks distance inner contour line

    outer_threshold : float (Default = 1.0)
        The threshold for the cooks distance outer contour line        

    kwargs : dict
        Keyword arguments that are passed to the base class and influence
        the visualization. Optional keyword arguments include:

        =========   ==========================================
        Property    Description
        --------    ------------------------------------------
        height      specify the height of the figure
        width       specify the width of the figure
        title       specify the title of the figure
        template    specify the template for the figure.
        =========   ==========================================    
    
    """

    def __init__(self, estimator, fig=None, inner_threshold=0.5, outer_threshold=1.0,
                **kwargs):
        super(ResidualsLeverage, self).__init__(estimator=estimator,
                                        fig=fig, **kwargs)

        self.inner_threshold = inner_threshold
        self.outer_threshold = outer_threshold
        self.title = self.title or str(estimator.description + \
            "<br>Residuals vs Leverage")

    def _cooks_contour(self, X, leverage, distance, sign):
        """Computes the leverage v standardized residuals, given Cooks Distance."""        
        # Compute F-distribution degrees of freedom
        n = X.shape[0]
        p = X.shape[1]

        # Designate the leverage range being plotted
        min_leverage = np.min(leverage)
        max_leverage = np.max(leverage)
        leverage_range = np.linspace(start=min_leverage, stop=max_leverage, num=n)

        # Compute standard residuals
        std_resid = sign * np.sqrt(distance * ((1-leverage_range)/leverage_range) * (p+1))        

        return leverage_range, std_resid         

    def _get_influence_points(self, cooks, distance):
        """Gets indices for influential points."""
        points = np.argwhere(np.greater(cooks,distance)).ravel()
        return points

    def fit(self, X, y):
        """Generates the plot.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training vector, where n_samples is the number of samples and
            n_features is the number of features.

        y : array-like, shape (n_samples) or (n_samples, n_features), optional
            Target relative to X for classification or regression;
            None for unsupervised learning.

        """
        # Compute standardized residuals
        self.estimator.fit(X,y)
        y_pred = self.estimator.predict(X)
        residuals = y - y_pred        
        residuals = standardized_residuals(X, y, y_pred)

        # Flatten arrays (just in case)
        y = y.ravel()
        y_pred = y_pred.ravel()
        residuals = residuals.ravel()

        # Compute Leverage and cooks distances
        lev = leverage(X) 
        cooks = cooks_distance(X, y, y_pred)

        # Create lowess smoothing line
        z1 = lowess(residuals, lev, frac=1./3, it=0, is_sorted=False, 
                    return_sorted=True)

        # Draw Cooks Distance contour lines 
        l1u_leverage, l1u_residual = \
            self._cooks_contour(X, lev, distance=self.inner_threshold, sign=1)
        l1l_leverage, l1l_residual = \
            self._cooks_contour(X, lev, distance=self.inner_threshold, sign=-1)
        l2u_leverage, l2u_residual = \
            self._cooks_contour(X, lev, distance=self.outer_threshold, sign=1)
        l2l_leverage, l2l_residual = \
            self._cooks_contour(X, lev, distance=self.outer_threshold, sign=-1)

        # Grab influential points
        influential_points = []
        distances = [self.inner_threshold, self.outer_threshold]
        for distance in distances:
            influential_points.extend(self._get_influence_points(cooks, distance))
        non_influential_points = [point for point in np.arange(len(residuals)) if \
            point not in influential_points]

        # Create scatterplot traces
        data = [
            go.Scatter(x=lev[non_influential_points], 
                        y=residuals[non_influential_points],
                        mode='markers+text',
                        marker=dict(color=COLORS['blue']),
                        name="Residual vs Leverage",
                        textposition="top center",
                        showlegend=False),
            go.Scattergl(x=lev[influential_points], 
                        y=residuals[influential_points], 
                        mode='markers+text',
                        marker=dict(color='red'),                        
                        text=influential_points,
                        name="High Influence Points",
                        textposition='top center',
                        showlegend=True),  
            go.Scattergl(x=z1[:,0], y=z1[:,1],
                        mode='lines',
                        marker=dict(color='red'),
                        name="Training Set Lowess",
                        showlegend=False),
            go.Scatter(x=l1u_leverage,
                       y=l1u_residual,
                       mode='lines',
                       line=dict(dash='dash'),
                       marker=dict(color='red'),
                       name="Cooks' D = " + str(self.inner_threshold),
                       showlegend=True),
            go.Scatter(x=l1l_leverage,
                       y=l1l_residual,
                       mode='lines',
                       line=dict(dash='dash'),
                       marker=dict(color='red'),
                       name="Cooks' D = 0.5 (Lower)",
                       showlegend=False),
            go.Scatter(x=l2u_leverage,
                       y=l2u_residual,
                       mode='lines',
                       line=dict(dash='dot'),
                       marker=dict(color='red'),
                       name="Cooks' D = " + str(self.outer_threshold),
                       showlegend=True),
            go.Scatter(x=l2l_leverage,
                       y=l2l_residual,
                       mode='lines',
                       line=dict(dash='dot'),
                       marker=dict(color='red'),
                       name="Cooks' D = 1 (Lower)",
                       showlegend=False)                                                                                                                                   
       ]

        # Compute x and y axis limits based 110% of the data range
        xmin = np.min(lev) + (0.1 * np.min(lev))
        xmax = np.max(lev) + (0.1 * np.max(lev)) 
        ymin = np.min(residuals) + (0.1 * np.min(residuals))
        ymax = np.max(residuals) + (0.1 * np.max(residuals))

        # Designate Layout
        layout = go.Layout(title=self.title, 
                        title_x=0.5,
                        height=self.height,
                        width=self.width,
                        xaxis_title="Leverage",
                        yaxis_title="Standardized Residuals",
                        xaxis=dict(zeroline=False, range=[xmin, xmax]),
                        yaxis=dict(zeroline=False, range=[ymin, ymax]),                       
                        showlegend=True,
                        legend=dict(bgcolor='white'),
                        template=self.template)

        # Create figure object
        self.fig = go.Figure(data=data, layout=layout)  
                                
# --------------------------------------------------------------------------  #
#                             COOKS DISTANCE                                  #
# --------------------------------------------------------------------------  #
class CooksDistance(ModelVisualatrix):        
    """Cooks Distance.

    Cook's distance" is a measure of the influence of each observation on the 
    regression coefficients. The Cook's distance statistic is a measure, 
    for each observation in turn, of the extent of change in model 
    estimates when that particular observation is omitted. Any observation 
    for which the Cook's distance is close to 0.5 or more, or that is 
    substantially larger than other Cook's distances 
    (highly influential data points), requires investigation.

    Parameters
    ----------
    fig : Plotly Figure or FigureWidget
        The plotting object. 

    estimator : MLStudio estimator object.
        The object that implements the 'fit' and 'predict' methods.

    threshold : float (Default - 0.5)
        The Cooks Distance threshold value for the threshold line.  

    kwargs : dict
        Keyword arguments that are passed to the base class and influence
        the visualization. Optional keyword arguments include:

        =========   ==========================================
        Property    Description
        --------    ------------------------------------------
        height      specify the height of the figure
        width       specify the width of the figure
        title       specify the title of the figure
        template    specify the template for the figure.
        =========   ==========================================    
    
    """

    def __init__(self, estimator, fig=None, threshold=0.5, **kwargs):
        super(CooksDistance, self).__init__(estimator=estimator,
                                        fig=fig, **kwargs)

        self.threshold = threshold
        self.title = self.title or estimator.description + \
            "<br>Cooks Distance" + "<br> Threshold = " + str(threshold)

    def fit(self, X, y):
        """Generates the plot.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training vector, where n_samples is the number of samples and
            n_features is the number of features.

        y : array-like, shape (n_samples) or (n_samples, n_features), optional
            Target relative to X for classification or regression;
            None for unsupervised learning.

        """
        # Compute predictions
        self.estimator.fit(X,y)
        y_pred = self.estimator.predict(X)

        # Flatten arrays (just in case)
        y = y.ravel()
        y_pred = y_pred.ravel()

        # Compute Cooks Distance
        cooks = cooks_distance(X, y, y_pred)

        # Grab influential points above threshold
        points = np.argwhere(np.greater(cooks,self.threshold)).ravel()        

        # Create scatterplot traces
        data = [
            go.Scattergl(x=np.arange(len(cooks)), 
                        y=[round(x,4) for x in cooks],
                        mode='lines',
                        marker=dict(color=COLORS['blue']),
                        name="Cooks Distance",                                                
                        showlegend=False)                                                                                                                        
       ]
        # Designate Layout
        layout = go.Layout(title=self.title, 
                        title_x=0.5,
                        height=self.height,
                        width=self.width,
                        xaxis_title="Observation",
                        yaxis_title="Cooks Distance",
                        showlegend=False,
                        template=self.template)

        # Create figure object
        self.fig = go.Figure(data=data, layout=layout)  

        # Add horizontal line if any appoach Cooks Distance of 0.5
        if len(cooks[cooks > self.threshold]) > 0:
           self.fig.add_shape(
               go.layout.Shape(
                   type="line",
                   x0=0,
                   y0=self.threshold,
                   x1=len(cooks),
                   y1=self.threshold,
                   line=dict(
                       color="darkgrey",
                       width=2
                   )
               )
           )
           self.fig.add_trace(
               go.Scatter(
                   x=points,
                   y=cooks[points],
                   mode='markers+text',
                   marker=dict(color=COLORS['blue']),
                   name='Influential Points',
                   text=points,
                   textposition='top center'
               )
           )                               


#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : ML Studio                                                         #
# Version : 0.1.0                                                             #
# File    : gradient_descent.py                                               #
# Python  : 3.8.3                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Wednesday, March 18th 2020, 4:34:57 am                      #
# Last Modified : Saturday, June 13th 2020, 9:52:07 pm                        #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
"""Gradient Descent Module"""
from abc import ABC, abstractmethod, ABCMeta
import sys
import copy
import warnings

import numpy as np
import pandas as pd
from scipy import sparse
from sklearn.base import BaseEstimator

from mlstudio.supervised.core.objectives import MSE, CrossEntropy, Adjiman
from mlstudio.supervised.core.objectives import CategoricalCrossEntropy
from mlstudio.supervised.core.optimizers import GradientDescentOptimizer
from mlstudio.supervised.core.regularizers import L2
from mlstudio.supervised.core.scorers import R2, Accuracy
from mlstudio.supervised.core.tasks import LinearRegression
from mlstudio.supervised.core.tasks import LogisticRegression
from mlstudio.supervised.core.tasks import MultinomialLogisticRegression
from mlstudio.supervised.observers.base import Observer, ObserverList
from mlstudio.supervised.observers.early_stop import Performance
from mlstudio.supervised.observers.history import BlackBox, Progress
from mlstudio.utils.data_manager import AddBiasTerm, unpack_parameters
from mlstudio.utils.data_manager import RegressionDataProcessor
from mlstudio.utils.data_manager import LogisticRegressionDataProcessor
from mlstudio.utils.data_manager import MulticlassDataProcessor, batch_iterator
from mlstudio.utils.validation import check_X, check_X_y, check_is_fitted
from mlstudio.visual.text import OptimizationReport
# =========================================================================== #
#                       GRADIENT DESCENT ABSTRACT                             #
# =========================================================================== #        
class GradientDescentAbstract(ABC,BaseEstimator):
    """Gradient Descent abstract base class."""

    def __init__(self, eta0=0.01, epochs=1000, 
                 theta_init=None, optimizer=None,  observers=None,
                 verbose=False, random_state=None):

        self.eta0 = eta0
        self.epochs = epochs
        self.theta_init = theta_init
        self.optimizer = optimizer 
        self.observers = observers
        self.verbose = verbose
        self.random_state = random_state

    # ----------------------------------------------------------------------- #
    @property    
    def description(self):
        """Returns the estimator description."""                         
        optimizer = self._optimizer.__class__.__name__       
        return 'Gradient Descent with ' + optimizer + ' Optimization'  

    @property
    def eta(self):
        return self._eta

    @eta.setter  
    def eta(self, x):
        self._eta = x
        
    @property
    def val_size(self):
        return self._val_size

    @val_size.setter
    def val_size(self, x):
        self._val_size = x

    @property
    def converged(self):
        return self._converged

    @converged.setter
    def converged(self, x):
        self._converged = x       

    @property
    def feature_names(self):
        return self._feature_names

    @feature_names.setter
    def feature_names(self, x):
        self._feature_names = x
    # ----------------------------------------------------------------------- #
    def _validate_params(self):
        """Performs parameter validation."""
        # Initial Learning Rate
        validate_range(param=self.eta0, minimum=0,
                       maximum=1, param_name='eta0',
                       left="open", right="open")        
        # Epochs
        validate_int(param=self.epochs, param_name='epochs')
        # Theta Init
        if self.theta_init:
            validate_array_like(param=self.theta_init, param_name='theta_init')
        # Optimizer
        if self.optimizer:
            validate_optimizer(self.optimizer)
        # Observers 
        if self.observers:
            validate_observers(self.observers)
        # Verbose 
        validate_bool(param=self.verbose, param_name='verbose')
        # Random State
        if self.random_state:
            validate_int(param=self.random_state, param_name='random_state')

    # ----------------------------------------------------------------------- #
    def _copy_mutable_parameters(self):
        """Makes deepcopies of mutable parameters and makes them private members."""
        # Copy observers e.g. Learning Rate, Early Stop
        self._observers = []
        if self.observers:
            for observer in self.observers:
                self._observers.append(copy.deepcopy(observer))     

        # The Optimizer algorithm
        if self.optimizer:
            self._optimizer = copy.deepcopy(self.optimizer)
        else:
            self._optimizer = GradientDescentOptimizer()

    # ----------------------------------------------------------------------- #
    def _create_observer_attributes(self):            
        self.blackbox_ = BlackBox()

    # ----------------------------------------------------------------------- #
    def _create_observer_list(self):
        """Adds all observers to the observer list that gets notified."""
        self._observers.append(self.blackbox_)
        # Add any additional default observers to observer dictionary
        if self.verbose:
            self._observers.append(Progress())

        # Create private list of observers
        self._observer_list = ObserverList()                
        for observer in self._observers:
            self._observer_list.append(observer)

        # Publish model parameters and instance on observer objects.
        self._observer_list.set_params(self.get_params())
        self._observer_list.set_model(self)            

    # ----------------------------------------------------------------------- #
    def _compile(self):        
        """Obtains, initializes object dependencies and registers observers."""

        self._set_objective()        
        self._copy_mutable_parameters()
        self._obtain_implicit_dependencies()
        self._create_observer_attributes()
        self._create_observer_list()

    # ----------------------------------------------------------------------- #
    def _on_train_begin(self, log=None):
        """Initializes all data, objects, and dependencies.
        
        Parameters
        ----------
        log : dict
            Data relevant this part of the process. 
        """
        log = log or {}
        self._compile()    
        self._epoch = 0      
        self._batch = 0 
        self._val_size = 0
        self._theta = None
        self._gradient = None
        self._current_state = {}
        self._converged = False    
        self._final_result = None
        self._best_result = None
        self._feature_names = None
        # Initialize learning rate
        self._eta = copy.copy(self.eta0)
        # Initialize training on observers
        self._observer_list.on_train_begin()
        # Prepares data and adds to estimator as attributes.
        if log:            
            self._prepare_data(log.get('X'), log.get('y'))
        # Weights are initialized based upon the number of features in the dataset 
        self._init_weights()

    # ----------------------------------------------------------------------- #
    def _on_train_end(self, log=None):
        """Finalizes training, formats attributes, and ensures object state is fitted.
        
        Parameters
        ----------
        log : dict
            Data relevant this part of the process. Currently not used, but 
            kept for future applications. 
        
        """
        log = log or {}
        self.n_iter_ = self._epoch         
        self._observer_list.on_train_end()
        self._get_results()
    # ----------------------------------------------------------------------- #
    def _on_epoch_begin(self, log=None):
        """Initializes all data, objects, and dependencies.
        
        Parameters
        ----------
        log : dict
            Data relevant this part of the process. Currently not used, but 
            kept for future applications. 
        """
        log = log or {}
        self._set_current_state()
        self._observer_list.on_epoch_begin(epoch=self._epoch, log=self._current_state)
    # ----------------------------------------------------------------------- #
    def _on_epoch_end(self, log=None):
        """Finalizes epoching, formats attributes, and ensures object state is fitted.
        
        Parameters
        ----------
        log : dict
            Data relevant this part of the process. Currently not used, but 
            kept for future applications. 
        
        """
        log = log or {}
        self._observer_list.on_epoch_end(epoch=self._epoch, log=self._current_state)
        self._epoch += 1

    # ----------------------------------------------------------------------- #            
    def _on_batch_begin(self, log=None):
        """Initializes the batch and notifies observers.
        
        Parameters
        ----------
        log : dict
            Data relevant this part of the process. Currently not used, but 
            kept for future applications. 
        
        """
        log = log or {}
        self._observer_list.on_batch_begin(batch=self._batch, log=log)        


    # ----------------------------------------------------------------------- #            
    def _on_batch_end(self, log=None):
        """Wraps up the batch and notifies observers.
        
        Parameters
        ----------
        log : dict
            Data relevant this part of the process. Currently not used, but 
            kept for future applications. 
        
        """
        log = log or {}
        self._observer_list.on_batch_end(batch=self._batch, log=log)            
        self._batch += 1 

    # ----------------------------------------------------------------------- #
    @abstractmethod
    def _set_objective(self):
        pass
    # ----------------------------------------------------------------------- #
    @abstractmethod
    def _obtain_implicit_dependencies(self):
        pass    
    # ----------------------------------------------------------------------- #
    @abstractmethod
    def _set_current_state(self):
        """Takes snapshot of current state and performance."""
        pass           
    # ----------------------------------------------------------------------- #
    def _prepare_data(self):
        pass
    # ----------------------------------------------------------------------- #
    @abstractmethod
    def _init_weights(self):
        pass    
    # ----------------------------------------------------------------------- #
    @abstractmethod
    def fit(self, X=None, y=None):
        pass    

    # ----------------------------------------------------------------------- #
    def _get_results(self):
        # Determine if best or final weights should be stored.
        borf = 'final'
        for observer in self._observers:
            if isinstance(observer, Performance):
                borf = observer.best_or_final
        
        # Format results
        if borf == 'best':
            self.final_theta_ = self._theta
            self.final_intercept_, self.final_coef_ = unpack_parameters(self.final_theta_)
            self.theta_ = self.best_results_.get('theta')
            self.intercept_, self.coef_ = unpack_parameters(self.theta_)
        else:
            self.theta_ = self._theta
            self.intercept_, self.coef_ = unpack_parameters(self.theta_)


# =========================================================================== #
#                    GRADIENT DESCENT PURE OPTIMIZER                          #
# =========================================================================== #
class GradientDescentPureOptimizer(GradientDescentAbstract):
    """Performs pure optimization of an objective function."""

    def __init__(self, eta0=0.01, epochs=1000, objective=None,
                 theta_init=None, optimizer=None,  observers=None, 
                 verbose=False, random_state=None):
        super(GradientDescentPureOptimizer, self).__init__(
            eta0 = eta0,
            epochs = epochs,
            theta_init = theta_init,
            optimizer = optimizer,
            observers = observers,
            verbose = verbose,
            random_state = random_state
        )
        self.objective = objective

    # ----------------------------------------------------------------------- #        
    def _set_objective(self):
        self._objective = self.objective

    # ----------------------------------------------------------------------- #        
    def _obtain_implicit_dependencies(self):
        pass

    # ----------------------------------------------------------------------- #
    def _init_weights(self):
        """Initializes parameters."""
        if self.theta_init is not None:
            if self.theta_init.shape[0] != 2:
                raise ValueError("Parameters theta must have shape (2,)")
            else:
                self._theta = self.theta_init
        else:            
            rng = np.random.RandomState(self.random_state)         
            self._theta = rng.randn(2)    

    # ----------------------------------------------------------------------- #
    def _set_current_state(self):
        """Takes snapshot of current state and performance."""        
        s = {}
        s['epoch'] = self._epoch
        s['eta'] = self._eta
        s['theta'] = self._theta
        s['train_cost'] = self._objective(self._theta)
        s['gradient'] = self._gradient
        s['gradient_norm'] = None
        if self._gradient is not None:
            s['gradient_norm'] = np.linalg.norm(self._gradient)
        self._current_state = s

    # ----------------------------------------------------------------------- #            
    def fit(self, X=None, y=None):
        """Performs the optimization of the objective function..
        
        Parameters
        ----------
        objective : object derived from Objective class
            The objective function to be optimized

        Returns
        -------
        self
        """
        
        self._on_train_begin()

        while (self._epoch < self.epochs and not self._converged):

            self._on_epoch_begin()

            cost = self._objective(self._theta)

            self._theta_new, self._gradient = self._optimizer(gradient=self._objective.gradient, \
                    learning_rate=self._eta, theta=copy.deepcopy(self._theta))                    

            self._on_epoch_end()
            

        self._on_train_end()
        return self   
        
    # ----------------------------------------------------------------------- #            
    def _get_results(self):
        self.theta_ = self._theta
        self.intercept_ = self._theta[0]
        self.coef_ = self._theta[1:]

        

# =========================================================================== #
#                        GRADIENT DESCENT ESTIMATOR                           #
# =========================================================================== # 
class GradientDescentEstimator(GradientDescentAbstract):
    """Gradient descent base class for all estimators.
    
    Performs gradient descent optimization to estimate the parameters theta
    that best fit the data.

    Parameters
    ----------
    eta0 : float
        The initial learning rate on open interval (0,1) 

    epochs : int
        The number of epochs to execute

    batch_size : None or int (default=None) 
        The number of observations to include in each batch. This also 
        specifies the gradient descent variant according to the following:

            Batch_Size      Variant
            ----------      -----------------------
            None            Batch Gradient Descent
            1               Stochastic Gradient Descent
            Other int       Minibatch Gradient Descent

    theta_init : array_like
        Contains the initial values for the parameters theta. Should include
        the bias parameter in addition to the feature parameters.

    optimizer : An Optimizer object or None
        The optimization algorithm to use. If None, the generic 
        GradientDescentOptimizer will be used.

    observers : list
        A list of observer objects.

    verbose : Bool or Int
        If False, the parameter is ignored. If an integer is provided, this 
        will be the number of epochs between progress reports.

    random_state : int or None
        If an int, this will be the random state used anywhere pseudo-randomization
        occurs.
    
    """
    def __init__(self, eta0=0.01, epochs=1000, batch_size=None, 
                 theta_init=None, optimizer=None, observers=None, 
                 verbose=False, random_state=None):
        super(GradientDescentEstimator, self).__init__(
            eta0 = eta0,
            epochs = epochs,
            theta_init = theta_init,
            optimizer = optimizer,
            observers = observers,
            verbose = verbose,
            random_state = random_state    
        )
        self.batch_size = batch_size               

    # ----------------------------------------------------------------------- #                
    @property
    def task(self):
        return self._task

    # ----------------------------------------------------------------------- #                
    @property
    def variant(self):
        """Returns the gradient descent variant based upon the batch size."""
        if self.batch_size is None:
            variant = "Batch Gradient Descent"
        elif self.batch_size == 1:
            variant = "Stochastic Gradient Descent"   
        else:
            variant = "Minibatch Gradient Descent"   
        return variant

    # ----------------------------------------------------------------------- #                
    @property
    def description(self):
        """Returns the estimator description."""                   
        task = self._task.name 

        # Optimizer Title
        optimizer_title = ""       
        if not isinstance(self._optimizer, GradientDescentOptimizer):
            optimizer_title = " (" + self._optimizer.__class__.__name__\
                + " Optimization)"
        
        # Regularizer Title        
        try: 
            regularizer_title = " with " + self._objective.regularizer_name
        except:
            regularizer_title = ""
        
        return task + ' by ' + self.variant + optimizer_title + \
            regularizer_title

    # ----------------------------------------------------------------------- #    
    def _prepare_data(self, X, y):
        """Prepares X and y data for training.
        
        X and y data is prepared and if a Performance observer with a 
        validation set size parameter is set, the data is split. The 
        data is then added to the estimator as attributes.

        """
        data = self._data_processor.fit_transform(X, y)
        # Set attributes from data.
        for k, v in data.items():     
            setattr(self, k, v)
            # Attempt to extract feature names from the 'X' array  
            if np.ndim(v) > 1:
                if v.shape[1] > 1:
                    try:
                        self.features_ =  v.dtype.names                     
                    except:
                        self.features_ = None  

    # ----------------------------------------------------------------------- #
    def _init_weights(self):
        """Initializes parameters."""
        if self.theta_init is not None:
            assert self.theta_init.shape == (self.X_train_.shape[1],), \
                "Initial parameters theta must have shape (n_features+1,)."
            self._theta = self.theta_init
            self._theta
        else:
            # Random normal initialization for weights.
            rng = np.random.RandomState(self.random_state)                
            self._theta = rng.randn(self.X_train_.shape[1]) 
            # Set the bias initialization to zero
            self._theta[0] = 0
            self._theta = self._theta

    # ----------------------------------------------------------------------- #
    def _set_current_state(self):
        """Takes snapshot of current state and performance."""
        s= {}
        s['epoch'] = self._epoch      
        s['eta'] = self._eta    
        s['theta'] = self._theta 
        
        # Compute training costs 
        y_out = self._task.compute_output(self._theta, self.X_train_)
        s['train_cost'] = self._objective(self._theta, self.y_train_, y_out)
        # If there is a scoring object, compute scores
        if self._scorer:
            y_pred = self._task.predict(self._theta, self.X_train_)
            s['train_score'] = self._scorer(self.y_train_, y_pred)

        # If we have a validation set, compute validation error and score
        if hasattr(self, 'X_val_'):
            if self.X_val_.shape[0] > 0:
                y_out_val = self._task.compute_output(self._theta, self.X_val_)
                s['val_cost'] = self._objective(self._theta, self.y_val_, y_out_val)        
                if self._scorer:
                    y_pred_val = self._task.predict(self._theta, self.X_val_)
                    s['val_score'] = self._scorer(self.y_val_, y_pred_val)

        # Grab Gradient. Note: 1st iteration, the gradient will be None
        s['gradient'] = self._gradient
        # Compute the gradient norm if not first iteration
        s['gradient_norm'] = None
        if self._gradient is not None:
            s['gradient_norm'] = np.linalg.norm(self._gradient) 
        # This reflects current state for the epoch sent to all observers.
        self._current_state = s
    # ----------------------------------------------------------------------- #    
    def fit(self, X, y):
        """Trains model until stop condition is met.
        
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data
        y : numpy array, shape (n_samples,)
            Target values 
        Returns
        -------
        self : returns instance of self
        """        
        train_log = {'X': X, 'y': y}
        self._on_train_begin(train_log)        

        while (self._epoch < self.epochs and not self._converged):            

            self._on_epoch_begin()

            for X_batch, y_batch in batch_iterator(self.X_train_, self.y_train_, batch_size=self.batch_size):

                self._on_batch_begin()
                
                # Compute model output
                y_out = self._task.compute_output(self._theta, X_batch)     

                # Compute costs
                cost = self._objective(self._theta, y_batch, y_out)

                # Format batch log
                log = {'batch': self._batch,'theta': self._theta, 
                       'train_cost': cost}

                # Compute gradient and update parameters 
                self._theta, self._gradient = self._optimizer(gradient=self._objective.gradient, \
                    learning_rate=self._eta, theta=copy.copy(self._theta),  X=X_batch, y=y_batch,\
                        y_out=y_out)                       

                # Update batch log
                log['gradient'] = self._gradient
                log['gradient_norm'] = np.linalg.norm(self._gradient) 
                self._on_batch_end(log=log)

            # Wrap up epoch
            self._on_epoch_end()

        self._on_train_end()
        return self 

    # ----------------------------------------------------------------------- #    
    def predict(self, X):
        """Computes prediction on test data.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input data
        
        Returns
        -------
        y_pred : prediction
        """
        check_is_fitted(self)
        X = check_X(X)
        X = AddBiasTerm().fit_transform(X)
        return self._task.predict(self.theta_, X)    

    # ----------------------------------------------------------------------- #    
    def _score(self, X, y):
        """Calculates scores during as the beginning of each training epoch."""        
        y_pred = self._task.predict(self.theta_, X)
        if self._scorer is None:
            raise Exception("Unable to compute score. No scorer object provided.")
        return self._scorer(y, y_pred)

    # ----------------------------------------------------------------------- #    
    def score(self, X, y):
        """Computes scores for test data after training.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input data
        
        y : array_like of shape (n_samples,) 
            The target variable.
        
        Returns
        -------
        score based upon the scorer object.
        
        """
        y_pred = self.predict(X)        
        if self._scorer is None:
            raise Exception("Unable to compute score. No scorer object provided.")        
        return self._scorer(y, y_pred)    

    # ----------------------------------------------------------------------- #    
    def summary(self, reports=None):  
        """Prints and optimization report. 

        Parameters
        ----------
        reports : list default=['summary', 'performance', 'critical_points',
                                'features', 'hyperparameters']
            The reports in the order to be rendered. The valid report names are:
                'summary' : prints summary data for optimzation
                'hyperparameters' : prints the hyperparameters used for the optimization
                'performance' : prints performance in terms of cost and scores
                'critical_points' : prints cost and scores at critical points 
                    during the optimization
                'features' : prints the best or final intercept and coeficients 
                    by feature name if feature names are available. Best results 
                    are printed if the Performance observer is used and the 
                    'best_or_final' parameter = 'best'. Otherwise, final results
                    will be printed.

        features : list of str
            A list containing the names of the features in the data set. 
        """
        optimization = OptimizationReport(model=self, reports=reports)
        optimization.report()
                    
# =========================================================================== #
#                        GRADIENT DESCENT REGRESSOR                           #
# =========================================================================== # 
class GradientDescentRegressor(GradientDescentEstimator):
    """Gradient descent base class for all estimators."""
    def __init__(self, eta0=0.01, epochs=1000, batch_size=None, 
                 objective=None, theta_init=None, optimizer=None,  
                 scorer=None, observers=None, verbose=False, 
                 random_state=None):
        super(GradientDescentRegressor, self).__init__(
            eta0 = eta0,
            epochs = epochs,
            theta_init = theta_init,
            optimizer = optimizer,            
            observers = observers,
            verbose = verbose,
            random_state = random_state,    
            batch_size=batch_size
        )
        self.scorer = scorer 
        self.objective  = objective

    # ----------------------------------------------------------------------- #
    def _set_objective(self):        
        """Defaults to mean squared error loss with L2 regularization."""
        self._objective = self.objective if self.objective else \
            MSE()

    # ----------------------------------------------------------------------- #    
    def _copy_mutable_parameters(self, log=None):
        super(GradientDescentEstimator, self)._copy_mutable_parameters()
        self._scorer = copy.deepcopy(self.scorer) if self.scorer else \
            R2()

    # ----------------------------------------------------------------------- #    
    def _obtain_implicit_dependencies(self):
        super(GradientDescentRegressor, self)._obtain_implicit_dependencies()                
        # Set the task that will be computing the output and predictions. 
        self._task = LinearRegression()

        # Instantiates the regression data processor for the 
        # _prepare_data method 
        self._data_processor = RegressionDataProcessor(estimator=self,
                                random_state=self.random_state)          


        


#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# ============================================================================ #
# Project : MLStudio                                                           #
# Version : 0.1.0                                                              #
# File    : scorers.py                                                         #
# Python  : 3.8.2                                                              #
# ---------------------------------------------------------------------------- #
# Author  : John James                                                         #
# Company : DecisionScients                                                    #
# Email   : jjames@decisionscients.com                                         #
# URL     : https://github.com/decisionscients/MLStudio                        #
# ---------------------------------------------------------------------------- #
# Created       : Wednesday, March 18th 2020, 4:34:57 am                       #
# Last Modified : Wednesday, March 18th 2020, 2:07:20 pm                       #
# Modified By   : John James (jjames@decisionscients.com)                      #
# ---------------------------------------------------------------------------- #
# License : BSD                                                                #
# Copyright (c) 2020 DecisionScients                                           #
# ============================================================================ #
"""Classification and regression scorer classes."""
from abc import ABC, abstractmethod
import math
import numpy as np
from sklearn.base import BaseEstimator

class Scorer(ABC, BaseEstimator):
    """Abstract base class for all metrics."""

    @abstractmethod
    def __init__(self):
        raise NotImplementedError("This method is not implemented for "
                                  "this Abstract Base Class.")

    @abstractmethod
    def __call__(self, y, y_pred):
        raise NotImplementedError("This method is not implemented for "
                                  "this Abstract Base Class.")

class RegressionScorer(Scorer):
    """Base class for regression metrics."""

    @abstractmethod
    def __init__(self):
        raise NotImplementedError("This method is not implemented for "
                                  "this Abstract Base Class.")

    @abstractmethod
    def __call__(self, y, y_pred):
        raise NotImplementedError("This method is not implemented for "
                                  "this Abstract Base Class.")

class ClassificationScorer(Scorer):
    """Base class for classification metrics."""

    @abstractmethod
    def __init__(self):
        raise NotImplementedError("This method is not implemented for "
                                  "this Abstract Base Class.")

    @abstractmethod
    def __call__(self, y, y_pred):
        raise NotImplementedError("This method is not implemented for "
                                  "this Abstract Base Class.")

# --------------------------------------------------------------------------- #
#                           REGRESSION METRICS                                #
# --------------------------------------------------------------------------- #
class SSR(RegressionScorer):
    """Computes sum squared residuals given"""

    def __init__(self):
        self.mode = 'min'
        self.name = 'residual_sum_squared_error'
        self.label = "Residual Sum Squared Error"
        self.stateful = False
        self.best = np.min
        self.better = np.less
        self.worst = np.Inf
        self.epsilon_factor = -1

    
    def __call__(self, y, y_pred):
        e = y - y_pred
        return np.sum(e**2)  

class SST(RegressionScorer):
    """Computes total sum of squares"""

    def __init__(self):
        self.mode = 'min'
        self.name = 'total_sum_squared_error'
        self.label = "Total Sum Squared Error"
        self.stateful = False
        self.best = np.min
        self.better = np.less
        self.worst = np.Inf
        self.epsilon_factor = -1

    
    def __call__(self, y, y_pred):
        y_avg = np.mean(y)
        e = y-y_avg                
        return np.sum(e**2)

class R2(RegressionScorer):
    """Computes coefficient of determination."""

    def __init__(self):
        self.mode = 'max'        
        self.name = 'R2'
        self.label = r"$R^2$"
        self.stateful = False
        self.best = np.max
        self.better = np.greater
        self.worst = -np.Inf
        self.epsilon_factor = 1

    
    def __call__(self, y, y_pred):
        self._ssr = SSR()
        self._sst = SST()
        r2 = 1 - (self._ssr(y, y_pred)/self._sst(y, y_pred))        
        return r2

class VarExplained(RegressionScorer):
    """Computes proportion of variance explained."""

    def __init__(self):
        self.mode = 'max'
        self.name = 'percent_variance_explained'
        self.label = "Percent Variance Explained"
        self.stateful = False
        self.best = np.max
        self.better = np.greater
        self.worst = -np.Inf
        self.epsilon_factor = 1

    
    def __call__(self, y, y_pred):
        var_explained = 1 - (np.var(y-y_pred) / np.var(y))
        return var_explained                   

class MAE(RegressionScorer):
    """Computes mean absolute error given data and parameters."""

    def __init__(self):
        self.mode = 'min'
        self.name = 'mean_absolute_error'
        self.label = "Mean Absolute Error"
        self.stateful = False
        self.best = np.min
        self.better = np.less
        self.worst = np.Inf
        self.epsilon_factor = -1
    
    def __call__(self, y, y_pred):
        e = abs(y-y_pred)
        return np.mean(e)


class MSE(RegressionScorer):
    """Computes mean squared error given data and parameters."""

    def __init__(self):
        self.mode = 'min'
        self.name = 'mean_squared_error'
        self.label = "Mean Squared Error"
        self.stateful = False
        self.best = np.min
        self.better = np.less
        self.worst = np.Inf
        self.epsilon_factor = -1
    
    def __call__(self, y, y_pred):        
        e = y - y_pred
        return np.mean(e**2)

class NMSE(RegressionScorer):
    """Computes negative mean squared error given data and parameters."""

    def __init__(self):
        self.mode = 'max'
        self.name = 'neg_mean_squared_error'
        self.label = "Negative Mean Squared Error"
        self.stateful = False
        self.best = np.max
        self.better = np.greater
        self.worst = -np.Inf
        self.epsilon_factor = 1

    
    def __call__(self, y, y_pred):        
        e = y - y_pred
        return -np.mean(e**2)

class RMSE(RegressionScorer):
    """Computes root mean squared error given data and parameters."""

    def __init__(self):
        self.mode = 'min'
        self.name = 'root_mean_squared_error'
        self.label = "Root Mean Squared Error"
        self.stateful = False
        self.best = np.min
        self.better = np.less
        self.worst = np.Inf
        self.epsilon_factor = -1
    
    def __call__(self, y, y_pred):
        e = y-y_pred
        return np.sqrt(np.mean(e**2)) 

class NRMSE(RegressionScorer):
    """Computes negative root mean squared error given data and parameters."""

    def __init__(self):
        self.mode = 'max'
        self.name = 'neg_root_mean_squared_error'
        self.label = "Negative Root Mean Squared Error"
        self.stateful = False
        self.best = np.max
        self.better = np.greater
        self.worst = -np.Inf
        self.epsilon_factor = 1

    
    def __call__(self, y, y_pred):
        e = y-y_pred
        return -np.sqrt(np.mean(e**2))

class MSLE(RegressionScorer):
    """Computes mean squared log error given data and parameters."""

    def __init__(self):
        self.mode = 'min'
        self.name = 'mean_squared_log_error'
        self.label = "Mean Squared Log Error"
        self.stateful = False
        self.best = np.min
        self.better = np.less
        self.worst = np.Inf
        self.epsilon_factor = -1
    
    def __call__(self, y, y_pred):
        e = np.log(y+1)-np.log(y_pred+1)
        y = np.clip(y, 1e-15, 1-1e-15)    
        y_pred = np.clip(y_pred, 1e-15, 1-1e-15)    
        e = np.log(y)-np.log(y_pred)
        return np.mean(e**2)

class RMSLE(RegressionScorer):
    """Computes root mean squared log error given data and parameters."""

    def __init__(self):
        self.mode = 'min'
        self.name = 'root_mean_squared_log_error'
        self.label = "Root Mean Squared Log Error"
        self.stateful = False
        self.best = np.min
        self.better = np.less
        self.worst = np.Inf
        self.epsilon_factor = -1
    
    def __call__(self, y, y_pred):
        y = np.clip(y, 1e-15, 1-1e-15)    
        y_pred = np.clip(y_pred, 1e-15, 1-1e-15)    
        e = np.log(y)-np.log(y_pred)
        return np.sqrt(np.mean(e**2))

class MEDAE(RegressionScorer):
    """Computes median absolute error given data and parameters."""

    def __init__(self):
        self.mode = 'min'
        self.name = 'median_absolute_error'
        self.label = "Median Absolute Error"
        self.stateful = False
        self.best = np.min
        self.better = np.less
        self.worst = np.Inf
        self.epsilon_factor = -1
    
    def __call__(self, y, y_pred):        
        return np.median(np.abs(y_pred-y))

class MAPE(RegressionScorer):
    """Computes mean absolute percentage given data and parameters."""

    def __init__(self):
        self.mode = 'min'
        self.name = 'mean_absolute_percentage_error'
        self.label = "Mean Absolute Percentage Error"
        self.stateful = False
        self.best = np.min
        self.better = np.less
        self.worst = np.Inf
        self.epsilon_factor = -1
    
    def __call__(self, y, y_pred):        
        return 100*np.mean(np.abs((y-y_pred)/y))


# --------------------------------------------------------------------------- #
#                       CLASSIFICATION METRICS                                #
# --------------------------------------------------------------------------- #
class Accuracy(ClassificationScorer):
    """Computes accuracy."""

    def __init__(self):
        self.mode = 'max'
        self.name = 'accuracy'
        self.label = "Accuracy"
        self.stateful = False
        self.best = np.max
        self.better = np.greater
        self.worst = -np.Inf
        self.epsilon_factor = 1
    
    def __call__(self, y, y_pred):
        """Computes accuracy as correct over total."""
        # If scoring multinomial logistical regression with one-hot vectors,
        return np.sum(np.equal(y,y_pred)) / y.shape[0]



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# ============================================================================ #
# Project : MLStudio                                                           #
# Version : 0.1.0                                                              #
# File    : format.py                                                          #
# Python  : 3.8.2                                                              #
# ---------------------------------------------------------------------------- #
# Author  : John James                                                         #
# Company : DecisionScients                                                    #
# Email   : jjames@decisionscients.com                                         #
# URL     : https://github.com/decisionscients/MLStudio                        #
# ---------------------------------------------------------------------------- #
# Created       : Wednesday, March 18th 2020, 12:39:08 am                      #
# Last Modified : Wednesday, March 18th 2020, 12:39:21 am                      #
# Modified By   : John James (jjames@decisionscients.com)                      #
# ---------------------------------------------------------------------------- #
# License : BSD                                                                #
# Copyright (c) 2020 DecisionScients                                           #
# ============================================================================ #
"""Miscellaneous formatting functions."""
#%%
import re
import random
import string
import textwrap

def proper(s):
    """Strips then capitalizes each word in a string.""" 
    s = s.replace("-", " ").title()
    s = s.replace("_", " ").title()
    return s    

def snake(s):
    """Converts string to snake case suitable for filenames."""
    s = re.sub(r"[^a-zA-Z0-9._// ]+", '', s)
    s = re.sub(r'\s+', ' ', s).strip().lower()
    s = s.replace(" ", "_")
    pattern = '_' + '{2,}'
    s = re.sub(pattern, '_', s)
    return s

def format_text(x):
    x = " ".join(x.split())
    formatted = textwrap.fill(textwrap.dedent(x))
    return formatted        


# %%



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : MLStudio                                                          #
# Version : 0.1.0                                                             #
# File    : debugging.py                                                      #
# Python  : 3.8.2                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Monday, March 23rd 2020, 4:06:07 pm                         #
# Last Modified : Monday, March 23rd 2020, 4:35:57 pm                         #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
"""Call back that performs gradient checking."""
import copy
import numpy as np
import pandas as pd

from scipy.sparse import issparse
from mlstudio.supervised.observers.base import Observer
from mlstudio.supervised.core.objectives import Cost
from mlstudio.utils.validation import is_valid_array_size

# --------------------------------------------------------------------------- #
#                              GRADIENT CHECK                                 #
# --------------------------------------------------------------------------- #
class GradientCheck(Observer):
    """Performs gradient checking."""

    def __init__(self, iterations=50, epsilon=1e-4, verbose=False):
        super(GradientCheck, self).__init__()
        self.epsilon = epsilon
        self.iterations = iterations
        self.verbose = verbose
        self.name = "GradientCheck"
        
    def on_train_begin(self, log=None):        
        """Initializes gradient check parameters.
        
        Parameters
        ----------
        log : dict
            Contains no information
        """
        super(GradientCheck, self).on_train_begin()        
        self._n = 0
        self._iteration = []
        self._theta = []
        self._gradients = []
        self._approximations = []
        self._differences = []
        self._results = []    
        # Obtain a copy of the objective function and turn gradient_scaling off
        self._objective = copy.deepcopy(self.model.objective)
        self._objective.turn_off_gradient_scaling

    def _check_cost_functions(self, log):
        """Computes gradient and approximation for cost functions."""
        X = self.model.X_train_
        y = self.model.y_train_
        theta = log.get('theta')               

        grad_approx = []
        for i in np.arange(len(theta)):

            # Compute theta differentials
            theta_plus = theta.copy()
            theta_minus = theta.copy()
            theta_plus[i] = theta_plus[i] + self.epsilon
            theta_minus[i] = theta_minus[i] - self.epsilon
            # Compute associated costs
            y_pred = self.model.task.compute_output(theta_plus, X)
            J_plus = self._objective(theta_plus, y, y_pred)
            y_pred = self.model.task.compute_output(theta_minus, X)
            J_minus = self._objective(theta_minus, y, y_pred)

            # Estimate the gradient
            grad_approx_i = (J_plus - J_minus) / (2 * self.epsilon)         
            grad_approx.append(grad_approx_i)
        
        # Compute gradient via back-propagation
        y_pred = self.model.task.compute_output(theta, X)
        grad = self._objective.gradient(theta, X, y, y_pred)         
        return grad, grad_approx

    def _check_benchmark_functions(self, log):
        """Computes gradient and approximation for benchmark functions."""
        theta = log.get('theta')               
        grad_approx = []
        for i in np.arange(len(theta)):

            # Compute theta differentials
            theta_plus = theta.copy()
            theta_minus = theta.copy()
            theta_plus[i] = theta_plus[i] + self.epsilon
            theta_minus[i] = theta_minus[i] - self.epsilon
            # Compute associated costs            
            J_plus = self._objective(theta_plus)            
            J_minus = self._objective(theta_minus)
            # Estimate the gradient
            grad_approx_i = (J_plus - J_minus) / (2 * self.epsilon)         
            grad_approx.append(grad_approx_i)
        
        # Compute gradient via back-propagation        
        grad = self._objective.gradient(theta)         
        return grad, grad_approx        



    def on_epoch_end(self, epoch, log=None):
        """Checks gradient each self.iterations number of iterations.

        Parameters
        ----------
        epoch : int
            The current epoch number

        log : dict
            Dictionary containing training cost, (and if metric=score, 
            validation cost)  

        """        
        if log.get('epoch') % self.iterations == 0:                                         
            if isinstance(self._objective, Cost):
                grad, grad_approx = self._check_cost_functions(log)
            else:
                grad, grad_approx = self._check_benchmark_functions(log)               

            grad = np.array(grad)
            grad_approx = np.array(grad_approx)

            # Check norms and bail if gradients are rediculous.
            if is_valid_array_size(grad) and is_valid_array_size(grad_approx):
                # Evaluate
                numerator = np.linalg.norm(grad-grad_approx)
                denominator = np.linalg.norm(grad) + np.linalg.norm(grad_approx)
                difference = numerator / denominator
                result = difference < self.epsilon

                # Update results
                self._n += 1
                self._iteration.append(self._n)
                self._theta.append(log.get('theta'))
                self._gradients.append(grad)
                self._approximations.append(grad_approx)
                self._differences.append(difference)                
                self._results.append(result)
            else:
                return

    def _print(self, results):
        """Prints results."""
        print("\n","* ",40*"=", " *")        
        print("Gradient Checks for {desc}".format(desc=self.model.description))
        print("  Num Failures : {failures}".format(failures=results['failures']))
        print(" Num Successes : {successes}".format(successes=results['successes']))
        print(" Pct Successes : {pct}".format(pct=results['successes']/(self._n)*100))
        print("Avg Difference : {diff}".format(diff=results['avg_difference']))        

    def _print_errors(self):
        """Prints gradient approximation errors."""
        def condition(x): return x > self.epsilon
        problems = [idx for idx, element in enumerate(self._differences) if condition(element)]
        for p in problems:
            print("\n\n  Iteration: {i} Theta: {t}".format(i=str(p), \
                t=self._theta[p]))
            print("  Iteration: {i} Gradient: {g}".format(i=str(p), \
                g=self._gradients[p]))
            print("  Iteration: {i} Approximation {a}".format(i=str(p), \
                a=self._approximations[p]))
            print("  Iteration: {i} Difference {a}".format(i=str(p), \
                a=self._differences[p]))                
        msg = "Gradient check failed for " + self._objective.name 
        if hasattr(self._objective, 'regularization'):
            msg = msg + ' with ' + self._objective.regularization.name 
        raise Exception(msg)        

    def on_train_end(self, log=None):
        d = {"Iteration": self._iteration, "Theta": self._theta,
             "Difference": self._differences,
             "Result": self._results, "Gradient": self._gradients,
             "Approximation": self._approximations}
        df = pd.DataFrame(d)
        results = {}
        results['failures'] = len(df[df['Result']== False])
        results['successes'] = len(df[df['Result']== True])
        results['avg_difference'] = df['Difference'].mean(axis=0)
        if self.verbose:
            self._print(results)
        if self._n > 0:
            if results['successes'] / self._n * 100 < 100.0:
                self._print_errors()



        


#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# ============================================================================ #
# Project : MLStudio                                                           #
# Version : 0.1.0                                                              #
# File    : model_selection.py                                                 #
# Python  : 3.8.2                                                              #
# ---------------------------------------------------------------------------- #
# Author  : John James                                                         #
# Company : DecisionScients                                                    #
# Email   : jjames@decisionscients.com                                         #
# URL     : https://github.com/decisionscients/MLStudio                        #
# ---------------------------------------------------------------------------- #
# Created       : Tuesday, March 17th 2020, 7:25:56 pm                         #
# Last Modified : Tuesday, March 17th 2020, 7:25:56 pm                         #
# Modified By   : John James (jjames@decisionscients.com)                      #
# ---------------------------------------------------------------------------- #
# License : BSD                                                                #
# Copyright (c) 2020 DecisionScients                                           #
# ============================================================================ #
""" Model selection plots.

Model selection visualizations include: 

    * Cost Curves: Training costs by epoch.
    * Learning Curves: Training and validation scores by training set sizes.
    * Scalability Curves: Fit times by training examples
    * Performance Curve: Scores by fit times.
    * Validation Curves: Training and validation scores by parameter.
    * Validation Surfaces: Validation scores by two parameters.


"""
import math
import numpy as np
import pandas as pd
import plotly
import plotly.express as px
import plotly.graph_objects as go 
import plotly.offline as py
from sklearn.model_selection import ParameterGrid, learning_curve 
from sklearn.model_selection import validation_curve
from sklearn.utils.validation import check_X_y, check_array

from .base import Visualatrix
from mlstudio.utils.format import proper
from mlstudio.visual.base import ModelVisualatrix
# ---------------------------------------------------------------------------- #
#                              COST CURVES                                     #
# ---------------------------------------------------------------------------- #
class CostCurve(ModelVisualatrix):
    """ Plots training costs by epoch.

    This visualization illuminates the path towards gradient descent convergence 
    over a designated number of epochs or until a stop condition is met. The
    basic line plot presents the number of epochs on the x-axis and costs on the
    y-axis. A factor variable may be provided to show multiple lines, one
    for each value of the factor variable. Subplots can also be created
    for each value of a subplot variable.

    Parameters
    ----------    
    estimator : MLStudio estimator object.
        The object that implements the 'fit' and 'predict' methods.

    fig : Plotly Figure or FigureWidget object.
        The object being evaluated.

    param_grid : dict (default=None)
        A dictionary in which the keys are estimator hyperparameter names, 
        and the values are hyperparameter values to be estimated. 
    
    factor : str or None (Default=None)
        The column to be used as a factor variable which groups data for 
        plotting.

    facet_col : str or None (Default=None)
        The param_grid key that is used to assign marks ot facetted subplots
        in the horizontal direction.  

    facet_col_wrap : int (Default=2)
        The maximum number of facet columns.  Wraps the column variable 
        at this width, so that the column facets span multiple rows.        

    kwargs : dict
        Keyword arguments that are passed to the base class and influence
        the visualization. Optional keyword arguments include:

        =========   ==========================================
        Property    Description
        --------    ------------------------------------------
        height      specify the height of the figure
        width       specify the width of the figure
        title       specify the title of the figure
        template    specify the template for the figure.
        =========   ==========================================


    """
    def __init__(self, estimator, fig=None, param_grid=None, 
                 factor=None, facet_col=None, facet_col_wrap=2, **kwargs):
        super(CostCurve, self).__init__(fig=fig, estimator=estimator, **kwargs)    

        self.param_grid = param_grid
        self.factor = factor
        self.facet_col = facet_col
        self.facet_col_wrap = facet_col_wrap
        self.title = self.title or str(estimator.description + \
            "<br>Cost Curve Plot")

    def fit(self, X, y=None):
        """Fits the model and creates the figure object.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training vector, where n_samples is the number of samples and
            n_features is the number of features.

        y : array-like, shape (n_samples) or (n_samples, n_features), optional
            Target relative to X for classification or regression;
            None for unsupervised learning.
        
        """
        self._validate(X, y)
        data = self._fit_plot_data(X, y)
        self.fig = self._fit_plot(data)

        return self


    def _validate(self, X, y):
        super(CostCurve, self)._validate(X,y)

        if self.param_grid:
            if not self.factor and not self.facet_col:
                raise ValueError("If using a param_grid, the color or facet_col\
                    parameters must be set.")
            if self.factor and self.factor not in self.param_grid:
                raise ValueError("The 'factor' parameter must be the 'param_grid'.")
            if self.facet_col and self.facet_col not in self.param_grid:
                raise ValueError("The 'facet_col' parameter must be the 'param_grid'.")
            if not isinstance(self.facet_col_wrap, int):
                raise ValueError("The 'facet_col_wrap' parameter must be an integer.")


    def _fit_plot_data(self, X, y):
        """Fits models and creates data in tidy format for plotting."""
        data = pd.DataFrame()

        # If a parameter grid was provided, set the parameters on the estimator.
        if self.param_grid:
            grid = ParameterGrid(self.param_grid)
            for params in grid:
                # Set params attributes on estimator
                d = {}
                for param, value in params.items():                    
                    setattr(self.estimator, param, value)
                    d[proper(param)] = round(value,4)

                self.estimator.fit(X,y)
                
                d['Epoch'] = np.arange(start=1, stop=self.estimator.blackbox_.total_epochs+1)
                d['Cost'] = self.estimator.blackbox_.epoch_log['train_cost']
                df = pd.DataFrame(d)
                data = pd.concat([data, df], axis=0)  

        # Otherwise, fit the model with current parameters as presented
        else:      
            self.estimator.fit(X,y)
            d = {}
            d['Epoch'] = np.arange(start=1, stop=self.estimator.blackbox_.total_epochs+1)
            d['Cost'] = self.estimator.blackbox_.epoch_log['train_cost']
            data = pd.DataFrame(d)

        return data
    
    def _fit_plot(self, data):
        """Creates the plot express object."""

        if self.factor and self.facet_col:
            fig = px.line(data, x='Epoch', y='Cost', color=proper(self.factor), 
                        facet_col=proper(self.facet_col),
                        facet_col_wrap=self.facet_col_wrap,
                        template=self.template,
                        title=self.title,
                        height=self.height,
                        width=self.width)        
        elif self.factor:
            fig = px.line(data, x='Epoch', y='Cost', color=proper(self.factor),
                        template=self.template,
                        title=self.title,
                        height=self.height,
                        width=self.width)

        elif self.facet_col:             
            fig = px.line(data, x='Epoch', y='Cost', 
                        facet_col=proper(self.facet_col),
                        facet_col_wrap=self.facet_col_wrap,
                        template=self.template,
                        title=self.title,
                        height=self.height,
                        width=self.width)                    
        else:
            fig = px.line(data, x='Epoch', y='Cost',
                        template=self.template,
                        title=self.title,
                        height=self.height,
                        width=self.width)            

        fig.update_layout(title_x=0.5)        
        return fig

# ---------------------------------------------------------------------------- #
#                              LEARNING CURVE                                  #
# ---------------------------------------------------------------------------- #        
class LearningCurve(ModelVisualatrix):        
    """Plots training and cross-validation scores by training sample size.

    Parameters
    ----------
    fig : Plotly Figure or FigureWidget object.
        The object being analyzed 

    estimator : MLStudio estimator object.
        The object that implements the 'fit' and 'predict' methods.

    cv : int, cross-validation generator or an iterable, optional
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:
        - None, to use the default 5-fold cross-validation,
        - integer, to specify the number of folds.
        - :term:`CV splitter`,
        - An iterable yielding (train, test) splits as arrays of indices.

        For integer/None inputs, if ``y`` is binary or multiclass,
        :class:`StratifiedKFold` used. If the estimator is not a classifier
        or if ``y`` is neither binary nor multiclass, :class:`KFold` is used.

        Refer :ref:`User Guide <cross_validation>` for the various
        cross-validators that can be used here.

    n_jobs : int or None, optional (default=None)
        Number of jobs to run in parallel.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.

    train_sizes : array-like, shape (n_ticks,), dtype float or int
        Relative or absolute numbers of training examples that will be used to
        generate the learning curve. If the dtype is float, it is regarded as a
        fraction of the maximum size of the training set (that is determined
        by the selected validation method), i.e. it has to be within (0, 1].
        Otherwise it is interpreted as absolute sizes of the training sets.
        Note that for classification the number of samples usually have to
        be big enough to contain at least one sample from each class.
        (default: np.linspace(0.1, 1.0, 5))
    
    kwargs : dict
        Keyword arguments that are passed to the base class and influence
        the visualization. Optional keyword arguments include:

        =========   ==========================================
        Property    Description
        --------    ------------------------------------------
        height      specify the height of the figure
        width       specify the width of the figure
        title       specify the title of the figure
        template    specify the template for the figure.
        =========   ==========================================    
    
    """

    def __init__(self, estimator, fig=None, cv=None, n_jobs=None, 
                 train_sizes=np.linspace(.1, 1.0, 5), **kwargs):
        super(LearningCurve, self).__init__(estimator=estimator, fig=fig, **kwargs)        

        self.cv = cv
        self.n_jobs = n_jobs
        self.train_sizes = train_sizes
        self.title = self.title or str(estimator.description + "<br>Learning Curve Plot")

    def fit(self, X, y):
        """Generates the learning curve plot

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training vector, where n_samples is the number of samples and
            n_features is the number of features.

        y : array-like, shape (n_samples) or (n_samples, n_features), optional
            Target relative to X for classification or regression;
            None for unsupervised learning.

        """
        train_sizes, train_scores, test_scores = \
            learning_curve(self.estimator, X, y, cv=self.cv, n_jobs=self.n_jobs,
                           train_sizes=self.train_sizes, return_times=False)

        # Extract statistics
        train_scores_mean = np.mean(train_scores, axis=1)
        train_scores_std = np.std(train_scores, axis=1)
        test_scores_mean = np.mean(test_scores, axis=1)
        test_scores_std = np.std(test_scores, axis=1)

        # Create confidence band by wrapping and reversing the data 
        # X line
        x = train_sizes
        x_rev = x[::-1]

        # Training Data
        train_upper = train_scores_mean + train_scores_std
        train_lower = train_scores_mean - train_scores_std
        train_lower = train_lower[::-1]

        # Validation Data
        test_upper = test_scores_mean + test_scores_std
        test_lower = test_scores_mean - test_scores_std
        test_lower = test_lower[::-1]        

        self.fig = go.Figure()

        # Plot training scores confidence bank
        self.fig.add_trace(go.Scatter(
            name='train_band',     
            mode='lines',       
            x=np.concatenate((x, x_rev), axis=0),
            y=np.concatenate((train_upper,train_lower), axis=0),
            fillcolor='#005b96',
            line_color='rgba(255,255,255,0)',
            fill='toself',
            opacity=0.15,
            showlegend=False
        ))

        self.fig.add_trace(go.Scatter(
            name='test_band',
            mode='lines',
            x=np.concatenate((x, x_rev), axis=0),
            y=np.concatenate((test_upper,test_lower), axis=0),
            fillcolor='rgb(27,158,119)',
            line_color='rgba(255,255,255,0)',
            fill='toself',
            opacity=0.15,
            showlegend=False
        ))

        # Plot training and validation lines
        self.fig.add_trace(go.Scatter(
            name='Training Scores',
            mode='lines+markers',
            x=x, y=train_scores_mean,
            line=dict(color='#005b96'),            
            showlegend=True
        ))

        self.fig.add_trace(go.Scatter(
            name='Cross-Validation Scores',
            mode='lines+markers',
            x=x, y=test_scores_mean,
            line=dict(color='rgb(27,158,119)'),             
            showlegend=True
        ))

        self.fig.update_layout(
            title=self.title,
            xaxis=dict(title='Training Samples'),
            yaxis=dict(title=self.estimator.scorer.label),
            title_x=0.5,
            template=self.template)

        
# ---------------------------------------------------------------------------- #
#                           MODEL SCALABILITY PLOT                             #
# ---------------------------------------------------------------------------- #        
class ModelScalability(ModelVisualatrix):        
    """Plots fit times against training samples.

    Parameters
    ----------
    fig : Plotly Figure or FigureWidget object.
        The object being analyzed 

    estimator : MLStudio estimator object.
        The object that implements the 'fit' and 'predict' methods.

    cv : int, cross-validation generator or an iterable, optional
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:
        - None, to use the default 5-fold cross-validation,
        - integer, to specify the number of folds.
        - :term:`CV splitter`,
        - An iterable yielding (train, test) splits as arrays of indices.

        For integer/None inputs, if ``y`` is binary or multiclass,
        :class:`StratifiedKFold` used. If the estimator is not a classifier
        or if ``y`` is neither binary nor multiclass, :class:`KFold` is used.

        Refer :ref:`User Guide <cross_validation>` for the various
        cross-validators that can be used here.

    n_jobs : int or None, optional (default=None)
        Number of jobs to run in parallel.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.


    train_sizes : array-like, shape (n_ticks,), dtype float or int
        Relative or absolute numbers of training examples that will be used to
        generate the learning curve. If the dtype is float, it is regarded as a
        fraction of the maximum size of the training set (that is determined
        by the selected validation method), i.e. it has to be within (0, 1].
        Otherwise it is interpreted as absolute sizes of the training sets.
        Note that for classification the number of samples usually have to
        be big enough to contain at least one sample from each class.
        (default: np.linspace(0.1, 1.0, 5))        
    
    kwargs : dict
        Keyword arguments that are passed to the base class and influence
        the visualization. Optional keyword arguments include:

        =========   ==========================================
        Property    Description
        --------    ------------------------------------------
        height      specify the height of the figure
        width       specify the width of the figure
        title       specify the title of the figure
        template    specify the template for the figure.
        =========   ==========================================    
    
    """

    def __init__(self, estimator, fig=None, cv=None, n_jobs=None, 
                 train_sizes=np.linspace(.1, 1.0, 5), **kwargs):
        super(ModelScalability, self).__init__(estimator=estimator, fig=fig, **kwargs)
        
        self.cv = cv
        self.n_jobs = n_jobs
        self.train_sizes = train_sizes
        self.title = self.title or str(estimator.description + "<br>Model Scalability Plot")

    def fit(self, X, y):
        """Generates the model scalability plot

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training vector, where n_samples is the number of samples and
            n_features is the number of features.

        y : array-like, shape (n_samples) or (n_samples, n_features), optional
            Target relative to X for classification or regression;
            None for unsupervised learning.

        """
        train_sizes, train_scores, test_scores, fit_times, _ = \
            learning_curve(self.estimator, X, y, cv=self.cv, n_jobs=self.n_jobs,
                           train_sizes=self.train_sizes, return_times=True)

        fit_times_mean = np.mean(fit_times, axis=1)
        fit_times_std = np.std(fit_times, axis=1)

        # Plot fit times by training samples
        fit_times_upper = go.Scatter(
            x=train_sizes, y=fit_times_mean + fit_times_std,
            mode='lines',
            marker=dict(color="#b3cde0"),            
            fillcolor="#b3cde0",
            fill='tonexty',
            showlegend=False
        )
        fit_times = go.Scatter(
            name='Fit Times',
            x=train_sizes, y=fit_times_mean, 
            mode='lines+markers',             
            line=dict(color='#005b96'),            
            marker=dict(color='#005b96'),
            fillcolor="#b3cde0",
            fill='tonexty',            
            showlegend=False
        )
        fit_times_lower = go.Scatter(
            x=train_sizes, y=fit_times_mean - fit_times_std,
            mode='lines',
            line=dict(color='#b3cde0'),            
            showlegend=False
        )        

        # Load from bottom up
        data = [fit_times_lower, fit_times, fit_times_upper]
        # Update layout with designated template
        layout = go.Layout(
            xaxis=dict(title='Training Samples'),
            yaxis=dict(title='Fit Times'),
            title=self.title,title_x=0.5,
            template=self.template
        )
        self.fig = go.Figure(data=data, layout=layout)
                    
# ---------------------------------------------------------------------------- #
#                        MODEL LEARNING PERFORMANCE                            #
# ---------------------------------------------------------------------------- #        
class ModelLearningPerformance(ModelVisualatrix):        
    """Plots scores against fit times.

    Parameters
    ----------
    fig : Plotly Figure or FigureWidget object.
        The object being analyzed 

    estimator : MLStudio estimator object.
        The object that implements the 'fit' and 'predict' methods.

    cv : int, cross-validation generator or an iterable, optional
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:
        - None, to use the default 5-fold cross-validation,
        - integer, to specify the number of folds.
        - :term:`CV splitter`,
        - An iterable yielding (train, test) splits as arrays of indices.

        For integer/None inputs, if ``y`` is binary or multiclass,
        :class:`StratifiedKFold` used. If the estimator is not a classifier
        or if ``y`` is neither binary nor multiclass, :class:`KFold` is used.

        Refer :ref:`User Guide <cross_validation>` for the various
        cross-validators that can be used here.

    n_jobs : int or None, optional (default=None)
        Number of jobs to run in parallel.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.


    train_sizes : array-like, shape (n_ticks,), dtype float or int
        Relative or absolute numbers of training examples that will be used to
        generate the learning curve. If the dtype is float, it is regarded as a
        fraction of the maximum size of the training set (that is determined
        by the selected validation method), i.e. it has to be within (0, 1].
        Otherwise it is interpreted as absolute sizes of the training sets.
        Note that for classification the number of samples usually have to
        be big enough to contain at least one sample from each class.
        (default: np.linspace(0.1, 1.0, 5))        
    
    kwargs : dict
        Keyword arguments that are passed to the base class and influence
        the visualization. Optional keyword arguments include:

        =========   ==========================================
        Property    Description
        --------    ------------------------------------------
        height      specify the height of the figure
        width       specify the width of the figure
        title       specify the title of the figure
        template    specify the template for the figure.
        =========   ==========================================    
    
    """

    def __init__(self, estimator, fig=None, cv=None, n_jobs=None, 
                 train_sizes=np.linspace(.1, 1.0, 5), **kwargs):
        super(ModelLearningPerformance, self).__init__(estimator=estimator,
                                                       fig=fig, **kwargs)

        self.cv = cv
        self.n_jobs = n_jobs
        self.train_sizes = train_sizes                                                       
        self.title = self.title or str(estimator.description + "<br>Model Learning Performance Plot")

    def fit(self, X, y):
        """Generates the model scalability plot

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training vector, where n_samples is the number of samples and
            n_features is the number of features.

        y : array-like, shape (n_samples) or (n_samples, n_features), optional
            Target relative to X for classification or regression;
            None for unsupervised learning.

        """
        train_sizes, train_scores, test_scores, fit_times, _ = \
            learning_curve(self.estimator, X, y, cv=self.cv, n_jobs=self.n_jobs,
                           train_sizes=self.train_sizes, return_times=True)

        fit_times_mean = np.mean(fit_times, axis=1)
        test_scores_mean = np.mean(test_scores, axis=1)
        test_scores_std = np.std(test_scores, axis=1)        

        # Plot fit times by training samples
        fit_times_upper = go.Scatter(
            x=fit_times_mean, y=test_scores_mean + test_scores_std,
            mode='lines',
            marker=dict(color="#b3cde0"),            
            fillcolor="#b3cde0",
            fill='tonexty',
            showlegend=False
        )
        fit_times = go.Scatter(
            name='Test Scores',
            x=fit_times_mean, y=test_scores_mean, 
            mode='lines+markers',             
            line=dict(color='#005b96'),            
            marker=dict(color='#005b96'),
            fillcolor="#b3cde0",
            fill='tonexty',            
            showlegend=False
        )
        fit_times_lower = go.Scatter(
            x=fit_times_mean, y=test_scores_mean - test_scores_std,
            mode='lines',
            line=dict(color='#b3cde0'),            
            showlegend=False
        )        

        # Load from bottom up
        data = [fit_times_lower, fit_times, fit_times_upper]
        # Update layout with designated template
        layout = go.Layout(
            xaxis=dict(title='Fit Times'),
            yaxis=dict(title=self.estimator.scorer.label),
            title=self.title,title_x=0.5,
            template=self.template
        )
        self.fig = go.Figure(data=data, layout=layout)
                
# ---------------------------------------------------------------------------- #
#                             LEARNING CURVES                                  #
# ---------------------------------------------------------------------------- #        
class LearningCurves(ModelVisualatrix):        
    """Plots all learning curve plots for a series of models.
    
    This class renders the set of learning curve plots for a designated
    set of models. For each model, the following plots are rendered:

        * Learning Curve
        * Model Scalability Plot
        * Model Learning Performance Plot.

    Parameters
    ----------
    fig : Plotly Figure or FigureWidget object
        The object being analyzed 

    estimators : a list of MLStudio estimator objects.
        The objects that implement the 'fit' and 'predict' methods.

    cv : int, cross-validation generator or an iterable, optional
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:
        - None, to use the default 5-fold cross-validation,
        - integer, to specify the number of folds.
        - :term:`CV splitter`,
        - An iterable yielding (train, test) splits as arrays of indices.

        For integer/None inputs, if ``y`` is binary or multiclass,
        :class:`StratifiedKFold` used. If the estimator is not a classifier
        or if ``y`` is neither binary nor multiclass, :class:`KFold` is used.

        Refer :ref:`User Guide <cross_validation>` for the various
        cross-validators that can be used here.

    n_jobs : int or None, optional (default=None)
        Number of jobs to run in parallel.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.


    train_sizes : array-like, shape (n_ticks,), dtype float or int
        Relative or absolute numbers of training examples that will be used to
        generate the learning curve. If the dtype is float, it is regarded as a
        fraction of the maximum size of the training set (that is determined
        by the selected validation method), i.e. it has to be within (0, 1].
        Otherwise it is interpreted as absolute sizes of the training sets.
        Note that for classification the number of samples usually have to
        be big enough to contain at least one sample from each class.
        (default: np.linspace(0.1, 1.0, 5))        
    
    kwargs : dict
        Keyword arguments that are passed to the base class and influence
        the visualization. Optional keyword arguments include:

        =========   ==========================================
        Property    Description
        --------    ------------------------------------------
        height      specify the height of the figure
        width       specify the width of the figure
        title       specify the title of the figure
        template    specify the template for the figure.
        =========   ========================================== 
    """

    def __init__(self, estimators, fig=None, cv=None, n_jobs=None, 
                 train_sizes=np.linspace(.1, 1.0, 5), **kwargs):
        super(LearningCurves, self).__init__(estimator=None, fig=fig, **kwargs)
        
        self.estimators = estimators
        self.cv = cv
        self.n_jobs = n_jobs
        self.train_sizes = train_sizes     
        self.title = self.title or "Learning Curve Analysis"

    def fit(self, X, y):
        """Generates the learning curve analysis plot

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training vector, where n_samples is the number of samples and
            n_features is the number of features.

        y : array-like, shape (n_samples) or (n_samples, n_features), optional
            Target relative to X for classification or regression;
            None for unsupervised learning.

        """
        # Designate shape of subplots
        rows = len(self.estimators)
        cols = 3

        # Obtain subplot titles
        subtitles = []
        for estimator in self.estimators:
            subtitles.append(estimator.description + "<br>Learning Curve")
            subtitles.append(estimator.description + "<br>Model Scalability Plot")
            subtitles.append(estimator.description + "<br>Model Learning Performance")

        # Create subplots
        self.fig = plotly.tools.make_subplots(rows=rows, cols=cols,
                                               subplot_titles=subtitles,
                                               horizontal_spacing=0.3/cols,
                                               vertical_spacing = 0.3/rows)

        # Update font size for subplot titles.
        for i in self.fig['layout']['annotations']:
            i['font']['size'] = 12
        
        # Populate cells of subplot
        for i, estimator in enumerate(self.estimators):
            lc = LearningCurve(estimator, cv=self.cv, n_jobs=self.n_jobs, \
                train_sizes=self.train_sizes)
            ms = ModelScalability(estimator, cv=self.cv, n_jobs=self.n_jobs, \
                train_sizes=self.train_sizes)
            mp = ModelLearningPerformance(estimator, cv=self.cv, \
                n_jobs=self.n_jobs, train_sizes=self.train_sizes)
        
            # Fit the models
            lc.fit(X,y)
            ms.fit(X,y)
            mp.fit(X,y)
        
            # Extract traces from figures and append to subplots
            traces = lc.fig.data
            for j, trace in enumerate(traces):
                if i == 1 and j in [2,3]:
                    trace.showlegend = True 
                else:
                    trace.showlegend = False               
                self.fig.append_trace(trace, i+1, 1)
            traces = ms.fig.data
            for trace in traces:
                self.fig.append_trace(trace, i+1, 2)
            traces = mp.fig.data
            for trace in traces:
                self.fig.append_trace(trace, i+1, 3)                
        
            # Update subplot xaxis titles
            self.fig.update_xaxes(title_text="Training Samples", row=i+1, col=1)
            self.fig.update_xaxes(title_text="Training Samples", row=i+1, col=2)
            self.fig.update_xaxes(title_text="Fit Times", row=i+1, col=3)

            # Update subplot yaxis titles
            self.fig.update_yaxes(title_text=estimator.scorer.label, row=i+1, col=1)
            self.fig.update_yaxes(title_text="Fit Times", row=i+1, col=2)
            self.fig.update_yaxes(title_text=estimator.scorer.label, row=i+1, col=3)

        #  Add title and template to layout.
        self.fig.update_layout(title_text="Learning Curve Analysis",
                                template=self.template,
                                height=rows*400, width=1200,
                                title_x=0.5)

        # Update subplot title font size
        for i in self.fig['layout']['annotations']:
            i['font']['size'] = 12
# ---------------------------------------------------------------------------- #
#                            VALIDATION CURVE                                  #
# ---------------------------------------------------------------------------- #        
class ValidationCurve(ModelVisualatrix):        
    """Plots training and cross-validation scores by hyperparameter.

    Parameters
    ----------
    fig : Plotly Figure or FigureWidget object
        The object being analyzed

    estimator : MLStudio estimator object.
        The object that implements the 'fit' and 'predict' methods.

    param_name : str
        The parameter being evaluated

    param_range : array-like
        The range of the parameter being evaluated.

    n_jobs : int or None, optional (default=None)
        Number of jobs to run in parallel.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.

    scoring : str
        The metric used to evaluate performance.         
    
    kwargs : dict
        Keyword arguments that are passed to the base class and influence
        the visualization. Optional keyword arguments include:

        =========   ==========================================
        Property    Description
        --------    ------------------------------------------
        height      specify the height of the figure
        width       specify the width of the figure
        title       specify the title of the figure
        template    specify the template for the figure.
        =========   ==========================================    
    
    """

    def __init__(self, estimator, param_name, param_range, fig=None, 
                 cv=None, n_jobs=None, scoring="r2", **kwargs):
        super(ValidationCurve, self).__init__(estimator=estimator, 
                                              fig=fig, **kwargs)
        
        self.param_name = param_name
        self.param_range = param_range
        self.cv = cv
        self.n_jobs = n_jobs
        self.scoring = scoring                
        self.title = self.title or str(estimator.description + "<br>Validation Curve")

    def fit(self, X, y):
        """Generates the validation curve plot

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training vector, where n_samples is the number of samples and
            n_features is the number of features.

        y : array-like, shape (n_samples) or (n_samples, n_features), optional
            Target relative to X for classification or regression;
            None for unsupervised learning.

        """
        train_scores, test_scores = \
            validation_curve(self.estimator, X, y, 
            param_name=self.param_name,
            param_range=self.param_range, scoring=self.scoring, 
            n_jobs=self.n_jobs)

        train_scores_mean = np.mean(train_scores, axis=1)
        train_scores_std = np.std(train_scores, axis=1)
        test_scores_mean = np.mean(test_scores, axis=1)
        test_scores_std = np.std(test_scores, axis=1)

        # Training Data
        train_upper = train_scores_mean + train_scores_std
        train_lower = train_scores_mean - train_scores_std
        train_lower = train_lower[::-1]

        # Validation Data
        test_upper = test_scores_mean + test_scores_std
        test_lower = test_scores_mean - test_scores_std
        test_lower = test_lower[::-1]        

        self.fig = go.Figure()

        x = self.param_range
        x_rev = x[::-1]

        # Plot training scores confidence bank
        self.fig.add_trace(go.Scatter(
            name='train_band',     
            mode='lines',       
            x=np.concatenate((x, x_rev), axis=0),
            y=np.concatenate((train_upper,train_lower), axis=0),
            fillcolor='#005b96',
            line_color='rgba(255,255,255,0)',
            fill='toself',
            opacity=0.15,
            showlegend=False
        ))

        self.fig.add_trace(go.Scatter(
            name='test_band',
            mode='lines',
            x=np.concatenate((x, x_rev), axis=0),
            y=np.concatenate((test_upper,test_lower), axis=0),
            fillcolor='rgb(27,158,119)',
            line_color='rgba(255,255,255,0)',
            fill='toself',
            opacity=0.15,
            showlegend=False
        ))

        # Plot training and validation lines
        self.fig.add_trace(go.Scatter(
            name='Training Scores',
            mode='lines+markers',
            x=x, y=train_scores_mean,
            line=dict(color='#005b96'),            
            showlegend=True
        ))

        self.fig.add_trace(go.Scatter(
            name='Cross-Validation Scores',
            mode='lines+markers',
            x=x, y=test_scores_mean,
            line=dict(color='rgb(27,158,119)'),             
            showlegend=True
        ))

        self.fig.update_layout(
            title=self.title,
            xaxis=dict(title=proper(self.param_name)),
            yaxis=dict(title=proper(self.scoring)),
            title_x=0.5,
            template=self.template)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : ML Studio                                                         #
# Version : 0.1.0                                                             #
# File    : regression.py                                                     #
# Python  : 3.8.3                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Friday, April 10th 2020, 3:27:23 pm                         #
# Last Modified : Wednesday, June 10th 2020, 9:11:49 pm                       #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
#%%
from collections import OrderedDict
import os
from pathlib import Path
import sys
homedir = str(Path(__file__).parents[4])
demodir = str(Path(__file__).parents[2])
datadir = os.path.join(demodir, "data")
sys.path.append(homedir)

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from sklearn.datasets import make_regression
from sklearn.linear_model import SGDRegressor, LinearRegression
from sklearn.model_selection import cross_val_score
from sklearn.model_selection import KFold, GridSearchCV
from sklearn.pipeline import Pipeline

from mlstudio.supervised.machine_learning.gradient_descent import GradientDescentRegressor
from mlstudio.supervised.core.objectives import MSE
from mlstudio.supervised.core.regularizers import L1, L2, L1_L2
from mlstudio.supervised.core.scorers import R2
from mlstudio.supervised.observers.early_stop import Performance
from mlstudio.utils.data_manager import StandardScaler, data_split
from mlstudio.utils.file_manager import save_fig
def get_data():
    X, y = make_regression(n_samples=100, n_features=10, 
                           n_informative=5, bias=0.5, effective_rank=8, 
                           noise=1.0, random_state=5)
    scaler = StandardScaler()
    X = scaler.fit_transform(X)    
    return X, y

def get_sgdregressor_results(X,y, algorithm):
    params = {'penalty': ['l1', 'l2', 'elasticnet'], 
              'alpha': [0.001, 0.01, 0.1],
              'l1_ratio': [0.5, 0.25, 0.15],
              'random_state': [50],
              'learning_rate': ['constant', 'optimal', 'invscaling', 'adaptive'],
              'eta0': [0.1, 0.01, 0.001,]}
    estimator = SGDRegressor()
    clf = GridSearchCV(estimator, params, scoring='r2')
    clf.fit(X, y)
    # Create scores dataframe
    scores = OrderedDict()
    scores['Algorithm'] = algorithm
    scores[r'$R^2$']= [max(0,x) for x in clf.cv_results_['mean_test_score']]
    scores = pd.DataFrame.from_dict(scores) 
    # Create times dataframe
    times = OrderedDict()
    times['Algorithm'] = algorithm
    times['Fit Times (secs)']= clf.cv_results_['mean_fit_time']
    times = pd.DataFrame.from_dict(times)     
    return scores, times

def get_mlstudio_results(X, y, algorithm, batch_size):
    params = {'eta0': [0.1, 0.01, 0.001],              
              'observers': [[Performance(mode='active', metric='val_score', epsilon=0.01)]],
              'objective': [MSE(regularizer=L1(alpha=0.001)),
                            MSE(regularizer=L1(alpha=0.01)),
                            MSE(regularizer=L1(alpha=0.1)),
                            MSE(regularizer=L2(alpha=0.001)),
                            MSE(regularizer=L2(alpha=0.01)),
                            MSE(regularizer=L2(alpha=0.1)),                            
                            MSE(regularizer=L1_L2(alpha=0.001, ratio=0.5)),
                            MSE(regularizer=L1_L2(alpha=0.001, ratio=0.25)),
                            MSE(regularizer=L1_L2(alpha=0.001, ratio=0.15)),
                            MSE(regularizer=L1_L2(alpha=0.01, ratio=0.5)),
                            MSE(regularizer=L1_L2(alpha=0.01, ratio=0.25)),
                            MSE(regularizer=L1_L2(alpha=0.01, ratio=0.15)),
                            MSE(regularizer=L1_L2(alpha=0.1, ratio=0.5)),
                            MSE(regularizer=L1_L2(alpha=0.1, ratio=0.25)),
                            MSE(regularizer=L1_L2(alpha=0.1, ratio=0.15))],
              'epochs': [1000], 'batch_size': batch_size,
              'random_state' : [50]}
    estimator = GradientDescentRegressor()
    clf = GridSearchCV(estimator, params, scoring='r2')
    clf.fit(X,y)    
    # Create scores dataframe
    scores = OrderedDict()
    scores['Algorithm'] = algorithm
    scores[r'$R^2$']= [max(0,x) for x in clf.cv_results_['mean_test_score']]
    scores = pd.DataFrame.from_dict(scores) 
    # Create times dataframe
    times = OrderedDict()
    times['Algorithm'] = algorithm
    times['Fit Times (secs)']= clf.cv_results_['mean_fit_time']
    times = pd.DataFrame.from_dict(times)     
    return scores, times        

def get_results(X, y):
    """Trains batch, stochastic and minibatch gradient descent."""    

    scores_filename = "GridsearchCV Scores.csv"
    scores_filepath = os.path.join(datadir, scores_filename)
    times_filename = "GridsearchCV Times.csv"
    times_filepath = os.path.join(datadir, times_filename)

    if os.path.exists(scores_filepath):
        scores = pd.read_csv(scores_filepath)
        times = pd.read_csv(times_filepath)

    else:
        scores, times = get_sgdregressor_results(X, y,
                                                algorithm="SKLearn SGDRegressor")        
        scores_batch, times_batch = get_mlstudio_results(X, y, 
                                                algorithm='Batch Gradient Descent', 
                                                batch_size=[None])
        scores_stochastic, times_stochastic = get_mlstudio_results(X, y, 
                                                algorithm='Stochastic Gradient Descent',
                                                batch_size=[1])
        scores_minibatch, times_minibatch = get_mlstudio_results(X, y, 
                                                algorithm='Minibatch Gradient Descent',
                                                batch_size=[32, 64, 128])    
        # Concatenate dataframes
        scores = pd.concat([scores, scores_batch, scores_stochastic, scores_minibatch], axis=0)
        times = pd.concat([times, times_batch, times_stochastic, times_minibatch], axis=0)
        # Save results as csv files
        scores.to_csv(scores_filepath)
        times.to_csv(times_filepath)
    return scores, times

X, y = get_data()
scores, times = get_results(X, y)
fig, axes = plt.subplots(1,2, figsize=(10,5))
fig.suptitle("Comparison Scikit-Learn vs ML Studio")
sns.barplot(x='Algorithm', y=r'$R^2$', data=scores, ci='sd', 
            palette="Blues_d", ax=axes[0])
axes[0].set_title("Coefficient of Determination Scores")
sns.barplot(x='Algorithm', y='Fit Times (secs)', data=times, ci='sd', 
            palette="Blues_d", ax=axes[1]) 
axes[1].set_title("Fit Times (secs)")
for ax in fig.axes:     
    ax.set_xticklabels(ax.get_xticklabels(), rotation=30, ha='right')                   
plt.draw()
#%%



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : MLStudio                                                          #
# Version : 0.1.0                                                             #
# File    : regression.py                                                     #
# Python  : 3.8.2                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Wednesday, March 18th 2020, 4:34:57 am                      #
# Last Modified : Monday, March 23rd 2020, 10:31:37 am                        #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
"""Regression algorithms.

This class encapsulates the core behaviors for regression classes. Currently,
the following regression classes are supported.
    
    * Linear Regression
    * Lasso Regression
    * Ridge Regression
    * ElasticNet Regression

The core behaviors exposed for each class include:

    * predict : Predicts outputs as linear combination of inputs and weights.
    * compute_cost : Computes cost associated with predictions
    * compute_gradient : Computes the derivative of loss w.r.t. to weights

"""
import numpy as np
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils.validation import check_array, check_is_fitted, check_X_y

from mlstudio.supervised.core.scorers import R2
from mlstudio.utils.data_manager import StandardScaler

# --------------------------------------------------------------------------- #
#                        LINEAR REGRESSION (OLS)                              #
# --------------------------------------------------------------------------- # 
class OLSRegression(BaseEstimator, RegressorMixin):
    """Ordinary least squares closed form linear regression."""

    def __init__(self, scorer=R2()):
        self.scorer = scorer

    def fit(self, X, y):
        """Fits the linear regression ordinary least squares solution.
        
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data

        y : numpy array, shape (n_samples,)
            Target values 

        Returns
        -------
        self : returns instance of self._
        """
        # Validate input and target 
        X, y = check_X_y(X, y)

        # Add bias term of ones to feature matrix
        X = np.insert(X, 0, 1.0, axis=1) 

        # Obtain n samples nad comp0ute matrix rank
        n = X.shape[1]
        r = np.linalg.matrix_rank(X)        

        # Find matrix equivalent using singular value decomposition
        U, S, V = np.linalg.svd(X)

        # Derive D^+ from sigma        
        D_plus = np.zeros((X.shape[0], X.shape[1])).T
        D_plus[:S.shape[0], :S.shape[0]] = np.linalg.inv(np.diag(S))
        
        # Compute Moore-Penrose pseudoinverse of X
        X_plus = V.T.dot(D_plus).dot(U.T)

        # Weights are the dot product of X_plus and y
        theta = X_plus.dot(y)

        # Save solution in attributes        
        self.intercept_ = theta[0]
        self.coef_ = theta[1:]

        return self

    def predict(self, X):
        """Computes prediction.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input data

        Returns
        -------
        y_pred : prediction
        """        
        # Check if fit had been called
        check_is_fitted(self)

        # Input validation
        X = check_array(X)

        # Create prediction as linear combination of inputs and parameters
        y_pred  = self.intercept_ + X.dot(self.coef_)
        return y_pred

    def score(self, X, y):
        """Computes scores using the scorer parameter.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input data

        y : array_like of shape (n_samples,) or (n_samples, n_classes)
            The target variable.

        Returns
        -------
        score based upon the scorer object.
        
        """
        X, y = check_X_y(X,y)
        y_pred = self.predict(X)        
        score = self.scorer(y, y_pred)
        return score






#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : MLStudio                                                          #
# Version : 0.1.0                                                             #
# File    : model_evaluation.py                                               #
# Python  : 3.8.2                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Thursday, March 19th 2020, 9:14:48 am                       #
# Last Modified : Thursday, March 19th 2020, 9:14:48 am                       #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
"""Model Evaluation Plots. """
import math
import numpy as np
import pandas as pd
import plotly
import plotly.express as px
import plotly.graph_objects as go 
import plotly.offline as py
from sklearn.model_selection import ParameterGrid, learning_curve 
from sklearn.model_selection import validation_curve

from .base import ModelVisualatrix
from mlstudio.supervised.machine_learning.linear_regression import LinearRegression
from mlstudio.supervised.machine_learning.ols_regression import OLSRegression
from mlstudio.utils.format import proper

# ---------------------------------------------------------------------------- #
#                             PREDICTION ERROR                                 #
# ---------------------------------------------------------------------------- #        
class PredictionError(ModelVisualatrix):        
    """Plots actual target values against predicted values.

    Parameters
    ----------
    fig : Plotly Figure or FigureWidget
        The plotting object. 
        
    estimator : MLStudio estimator object.
        The object that implements the 'fit' and 'predict' methods.
    
    kwargs : dict
        Keyword arguments that are passed to the base class and influence
        the visualization. Optional keyword arguments include:

        =========   ==========================================
        Property    Description
        --------    ------------------------------------------
        height      specify the height of the figure
        width       specify the width of the figure
        title       specify the title of the figure
        template    specify the template for the figure.
        =========   ==========================================    
    
    """

    def __init__(self, estimator, fig=None, **kwargs):
        super(PredictionError, self).__init__(estimator=estimator, 
                                              fig=fig, **kwargs)
        self.title = self.title or str(estimator.description + "<br>Prediction Error")

    def fit(self, X, y):
        """Generates the prediction error plot.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training vector, where n_samples is the number of samples and
            n_features is the number of features.

        y : array-like, shape (n_samples) or (n_samples, n_features), optional
            Target relative to X for classification or regression;
            None for unsupervised learning.

        """
        # Compute predicted vs actual.
        self.estimator.fit(X,y)
        y_pred = self.estimator.predict(X)

        # Compute best fit line predicted vs actual
        y = y.reshape(-1,1)
        est = OLSRegression()
        est.fit(y, y_pred) 

        # Format data for identity and best fit lines
        y = y.ravel()
        best_fit_x = np.arange(min(y), max(y))
        best_fit_y = est.intercept_ + est.coef_ * best_fit_x
        identity_x = best_fit_x
        identity_y = best_fit_x 

        # Scatterplot of predicted vs actual
        scatter = go.Scatter(
            x=y, y=y_pred,
            mode='markers',
            marker=dict(color='#005b96'),
            line_color='rgba(255,255,255,0.5)',
            opacity=0.75,
            showlegend=False
        )

        # Plot best fit line
        best_fit = go.Scatter(
            name='Best Fit',
            x=best_fit_x, y=best_fit_y, 
            mode='lines',  
            line=dict(color='#005b96'),            
            showlegend=True
        )
        identity = go.Scatter(
            name='Identity',
            x=identity_x, y=identity_y,
            mode='lines',
            line=dict(color='#b3cde0'),            
            showlegend=True
        )        

        # Load from bottom up
        data = [scatter, best_fit, identity]
        # Update layout with designated template
        layout = go.Layout(
            xaxis=dict(title='y'),
            yaxis=dict(title=r'$\hat{y}$'),
            title=self.title,title_x=0.5,
            template=self.template
        )
        self.fig = go.Figure(data=data, layout=layout)




#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : ML Studio                                                         #
# Version : 0.1.0                                                             #
# File    : regression.py                                                     #
# Python  : 3.8.3                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Friday, April 10th 2020, 3:27:23 pm                         #
# Last Modified : Wednesday, June 10th 2020, 9:11:49 pm                       #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
#%%
from collections import OrderedDict
import os
from pathlib import Path
import sys
homedir = str(Path(__file__).parents[4])
demodir = str(Path(__file__).parents[2])
sys.path.append(homedir)

import pandas as pd
import numpy as np
from sklearn.linear_model import SGDRegressor

from mlstudio.supervised.machine_learning.gradient_descent import GradientDescentRegressor
from mlstudio.supervised.core.scorers import R2
from mlstudio.supervised.core.objectives import MSE
from mlstudio.utils.data_manager import StandardScaler, data_split
from mlstudio.visual.animations import animate_optimization_regression


def get_data():
    """Obtains the Ames housing price data for modeling."""
    # ----------------------------------------------------------------------  #
    # Designate file locations
    datadir = os.path.join(homedir,"mlstudio/demo/data/Ames/")
    filepath = os.path.join(datadir, "train.csv")
    # ----------------------------------------------------------------------  #
    # Obtain and scale data
    cols = ["GrLivArea", "SalePrice"]
    df = pd.read_csv(filepath, nrows=500, usecols=cols)
    df_samples = df.head()
    X = np.array(df['GrLivArea']).reshape(-1,1)
    y = df['SalePrice']
    X_train, X_test, y_train, y_test = data_split(X,y, random_state=5)
    scaler = StandardScaler()
    X_train = scaler.fit_transform(X_train)
    X_test = scaler.transform(X_test)
    return X_train, X_test, y_train, y_test

def train_models(X, y):
    """Trains batch, stochastic and minibatch gradient descent."""    
    bgd = GradientDescentRegressor(theta_init=np.array([0,0]), epochs=500, 
                                   random_state=50,scorer=R2(),
                                   objective=MSE())
    sgd = GradientDescentRegressor(theta_init=np.array([0,0]), epochs=500, 
                                   batch_size=1, random_state=50, scorer=R2(),
                                   objective=MSE())
    mbgd = GradientDescentRegressor(theta_init=np.array([0,0]), epochs=500, 
                                    batch_size=64, random_state=50, scorer=R2(),
                                    objective=MSE())
    bgd.fit(X,y)
    sgd.fit(X,y)
    mbgd.fit(X,y)
    estimators = {'Batch Gradient Descent': bgd, 'Stochastic Gradient Descent': sgd,
            'Minibatch Gradient Descent': mbgd}
    return estimators

def append_filepath(filepath=None, appendage=None):
    """Appends the appendage to the filepath."""
    if filepath:
        base = os.path.basename(filepath)
        ext = os.path.splitext(base)[1]
        base = base + "_" + appendage + ext
        filepath = os.path.join(os.path.dirname(filepath), base)
    return filepath

def plot_optimization(estimators, max_frames=None, filepath=None, show=True):
    """Renders surface, scatterplot, and line plots."""        
    filepath = append_filepath(filepath, "optimization")
    # Render plot
    animate_optimization_regression(estimators=estimators, max_frames=max_frames, 
                                    filepath=filepath, show=show)

def regression_demo(max_frames=None, filepath=None, show=True):
    """Regression demo main processing function.
    
    Parameters
    ----------
    filepath:  str
        A relative or absolute filepath. 

    show: bool (default=True)
        Indicates whether to render the plot  
    
    """
     
    X_train, X_test, y_train, y_test = get_data()
    estimators = train_models(X_train, y_train)
    plot_optimization(estimators=estimators, max_frames=max_frames, 
                      filepath=filepath, show=show)

def main():
    filepath = os.path.join(demodir, 'figures/regression_demo.html')
    regression_demo(max_frames=100, filepath=filepath)

if __name__ == "__main__":
    main()
#%%

